#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os

from qiling.exception import *
from qiling.os.windows.const import *

from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

def _GetModuleHandle(ql, address, params):
    lpModuleName = params["lpModuleName"]
    if lpModuleName == 0:
        ret = ql.loader.pe_image_address
    else:
        lpModuleName = lpModuleName.lower()
        if not is_file_library(lpModuleName):
            lpModuleName += ".dll"
        if lpModuleName in ql.loader.dlls:
            ret = ql.loader.dlls[lpModuleName]
        else:
            ql.log.debug("Library %s not imported" % lpModuleName)
            ret = 0
    return ret


# HMODULE GetModuleHandleA(
#   LPCSTR lpModuleName
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetModuleHandleA(ql, address, params):
    return _GetModuleHandle(ql, address, params)


# HMODULE GetModuleHandleW(
#   LPCWSTR lpModuleName
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetModuleHandleW(ql, address, params):
    return _GetModuleHandle(ql, address, params)


# BOOL GetModuleHandleExW(
#   DWORD   dwFlags,
#   LPCWSTR lpModuleName,
#   HMODULE *phModule
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetModuleHandleExW(ql, address, params):
    res = _GetModuleHandle(ql, address, params)
    dst = params["phModule"]
    ql.mem.write(dst, res.to_bytes(4, byteorder="little"))
    return 1


# DWORD GetModuleFileNameA(
#   HMODULE hModule,
#   LPSTR   lpFilename,
#   DWORD   nSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetModuleFileNameA(ql, address, params):
    ret = 0
    hModule = params["hModule"]
    lpFilename = params["lpFilename"]
    nSize = params["nSize"]

    # GetModuleHandle can return pe_image_address as handle, and GetModuleFileName will try to retrieve it.
    # Pretty much 0 and pe_image_address value should do the same operations
    if not ql.code and (hModule == 0 or hModule == ql.loader.pe_image_address):
        filename = ql.loader.filepath
        filename_len = len(filename)
        if filename_len > nSize - 1:
            filename = ql.loader.filepath[:nSize - 1]
            ret = nSize
        else:
            ret = filename_len
        ql.mem.write(lpFilename, filename + b"\x00")
    else:
        ql.log.debug("hModule %x" % hModule)
        raise QlErrorNotImplemented("API not implemented")
    return ret


# DWORD GetModuleFileNameW(
#   HMODULE hModule,
#   LPSTR   lpFilename,
#   DWORD   nSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetModuleFileNameW(ql, address, params):
    ret = 0
    hModule = params["hModule"]
    lpFilename = params["lpFilename"]
    nSize = params["nSize"]
    # GetModuleHandle can return pe_image_address as handle, and GetModuleFileName will try to retrieve it.
    # Pretty much 0 and pe_image_address value should do the same operations
    if not ql.code and (hModule == 0 or hModule == ql.loader.pe_image_address):
        filename = ql.loader.filepath.decode('ascii').encode('utf-16le')
        filename_len = len(filename)
        if filename_len > nSize - 1:
            filename = ql.loader.filepath[:nSize - 1]
            ret = nSize
        else:
            ret = filename_len
        ql.mem.write(lpFilename, filename + b"\x00")
    else:
        raise QlErrorNotImplemented("API not implemented")
    return ret


# FARPROC GetProcAddress(
#   HMODULE hModule,
#   LPCSTR  lpProcName
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCSTR': 'POINTER'})
def hook_GetProcAddress(ql, address, params):
    if params["lpProcName"] > MAXUSHORT:
        # Look up by name
        params["lpProcName"] = ql.os.read_cstring(params["lpProcName"])
        lpProcName = bytes(params["lpProcName"], "ascii")
    else:
        # Look up by ordinal
        lpProcName = params["lpProcName"]
    # TODO fix for gandcrab
    if params["lpProcName"] == "RtlComputeCrc32":
        return 0

    # Check if dll is loaded
    try:
        dll_name = [key for key, value in ql.loader.dlls.items() if value == params['hModule']][0]
    except IndexError as ie:
        ql.log.info('Failed to import function "%s" with handle 0x%X' % (lpProcName, params['hModule']))
        return 0

    # Handle case where module is self
    if dll_name == os.path.basename(ql.loader.path):
        for addr, export in ql.loader.export_symbols.items():
            if export['name'] == lpProcName:
                return addr

    if lpProcName in ql.loader.import_address_table[dll_name]:
        return ql.loader.import_address_table[dll_name][lpProcName]

    return 0


# HMODULE LoadLibraryA(
#   LPCSTR lpLibFileName
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_LoadLibraryA(ql, address, params):
    lpLibFileName = params["lpLibFileName"]
    if not ql.code and lpLibFileName == ql.loader.filepath.decode():
        # Loading self
        return ql.loader.pe_image_address
    dll_base = ql.loader.load_dll(lpLibFileName.encode())
    return dll_base


# HMODULE LoadLibraryExA(
#   LPCSTR lpLibFileName,
#   HANDLE hFile,
#   DWORD  dwFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HANDLE': 'POINTER'})
def hook_LoadLibraryExA(ql, address, params):
    lpLibFileName = params["lpLibFileName"]
    dll_base = ql.loader.load_dll(lpLibFileName.encode())
    return dll_base


# HMODULE LoadLibraryW(
#   LPCWSTR lpLibFileName
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_LoadLibraryW(ql, address, params):
    lpLibFileName = params["lpLibFileName"].encode()
    dll_base = ql.loader.load_dll(lpLibFileName)
    return dll_base


# HMODULE LoadLibraryExW(
#   LPCSTR lpLibFileName,
#   HANDLE hFile,
#   DWORD  dwFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HANDLE': 'POINTER'})
def hook_LoadLibraryExW(ql, address, params):
    lpLibFileName = params["lpLibFileName"].encode()
    dll_base = ql.loader.load_dll(lpLibFileName)
    return dll_base


# DWORD SizeofResource(
#   HMODULE hModule,
#   HRSRC   hResInfo
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HMODULE': 'POINTER'})
def hook_SizeofResource(ql, address, params):
    # Return size of resource
    # TODO set a valid value. More tests have to be made to find it.
    return 0x8


# HGLOBAL LoadResource(
#   HMODULE hModule,
#   HRSRC   hResInfo
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HMODULE': 'POINTER'})
def hook_LoadResource(ql, address, params):
    pointer = params["hResInfo"]
    return pointer


# LPVOID LockResource(
#   HGLOBAL hResData
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HMODULE': 'POINTER'})
def hook_LockResource(ql, address, params):
    pointer = params["hResData"]
    return pointer


# BOOL DisableThreadLibraryCalls(
#  HMODULE hLibModule
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HMODULE': 'POINTER'})
def hook_DisableThreadLibraryCalls(ql, address, params):
    return 1


# BOOL FreeLibrary(
#   HMODULE hLibModule
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HMODULE': 'POINTER'})
def hook_FreeLibrary(ql, address, params):
    return 1


# BOOL SetDefaultDllDirectories(
#   DWORD DirectoryFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_SetDefaultDllDirectories(ql, address, params):
    value = params["DirectoryFlags"]
    if value == LOAD_LIBRARY_SEARCH_USER_DIRS:
        # TODO we have to probably set an handler for this, since it can be a not default value.
        #  And we have to change the default path of load
        raise QlErrorNotImplemented("API not implemented")



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from unicorn.mips_const import *

reg_map = {
    "zero": UC_MIPS_REG_ZERO, 
    "at": UC_MIPS_REG_AT, 
    "v0": UC_MIPS_REG_V0,
    "v1": UC_MIPS_REG_V1, 
    "a0": UC_MIPS_REG_A0, 
    "a1": UC_MIPS_REG_A1,
    "a2": UC_MIPS_REG_A2, 
    "a3": UC_MIPS_REG_A3, 
    "t0": UC_MIPS_REG_T0,
    "t1": UC_MIPS_REG_T1, 
    "t2": UC_MIPS_REG_T2, 
    "t3": UC_MIPS_REG_T3,
    "t4": UC_MIPS_REG_T4, 
    "t5": UC_MIPS_REG_T5, 
    "t6": UC_MIPS_REG_T6,
    "t7": UC_MIPS_REG_T7, 
    "s0": UC_MIPS_REG_S0,
    "s1": UC_MIPS_REG_S1,
    "s2": UC_MIPS_REG_S2, 
    "s3": UC_MIPS_REG_S3, 
    "s4": UC_MIPS_REG_S4,
    "s5": UC_MIPS_REG_S5, 
    "s6": UC_MIPS_REG_S6, 
    "s7": UC_MIPS_REG_S7,
    "t8": UC_MIPS_REG_T8, 
    "t9": UC_MIPS_REG_T9, 
    "k0": UC_MIPS_REG_K0,
    "k1": UC_MIPS_REG_K1,
    "gp": UC_MIPS_REG_GP, 
    "sp": UC_MIPS_REG_SP,
    "s8": UC_MIPS_REG_S8,
    "ra": UC_MIPS_REG_RA,
    "status": UC_MIPS_REG_INVALID, 
    "lo": UC_MIPS_REG_LO, 
    "hi": UC_MIPS_REG_HI, 
    "badvaddr": UC_MIPS_REG_INVALID,
    "cause":UC_MIPS_REG_INVALID,
    "pc": UC_MIPS_REG_PC,
}

reg_map_afpr128 = {
    "cp0_config3" : UC_MIPS_REG_CP0_CONFIG3,
    "cp0_userlocal": UC_MIPS_REG_CP0_USERLOCAL,
}


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

EFI_MAX_BIT = (1 << 63)
EFI_SUCCESS = 0

EFI_LOAD_ERROR				= EFI_MAX_BIT |  1
EFI_INVALID_PARAMETER		= EFI_MAX_BIT |  2
EFI_UNSUPPORTED				= EFI_MAX_BIT |  3
EFI_BAD_BUFFER_SIZE			= EFI_MAX_BIT |  4
EFI_BUFFER_TOO_SMALL		= EFI_MAX_BIT |  5
EFI_NOT_READY				= EFI_MAX_BIT |  6
EFI_DEVICE_ERROR			= EFI_MAX_BIT |  7
EFI_WRITE_PROTECTED			= EFI_MAX_BIT |  8
EFI_OUT_OF_RESOURCES		= EFI_MAX_BIT |  9
EFI_VOLUME_CORRUPTED		= EFI_MAX_BIT | 10
EFI_VOLUME_FULL				= EFI_MAX_BIT | 11
EFI_NO_MEDIA				= EFI_MAX_BIT | 12
EFI_MEDIA_CHANGED			= EFI_MAX_BIT | 13
EFI_NOT_FOUND				= EFI_MAX_BIT | 14
EFI_ACCESS_DENIED			= EFI_MAX_BIT | 15
EFI_NO_RESPONSE				= EFI_MAX_BIT | 16
EFI_NO_MAPPING				= EFI_MAX_BIT | 17
EFI_TIMEOUT					= EFI_MAX_BIT | 18
EFI_NOT_STARTED				= EFI_MAX_BIT | 19
EFI_ALREADY_STARTED			= EFI_MAX_BIT | 20
EFI_ABORTED					= EFI_MAX_BIT | 21
EFI_ICMP_ERROR				= EFI_MAX_BIT | 22
EFI_TFTP_ERROR				= EFI_MAX_BIT | 23
EFI_PROTOCOL_ERROR			= EFI_MAX_BIT | 24
EFI_INCOMPATIBLE_VERSION	= EFI_MAX_BIT | 25
EFI_SECURITY_VIOLATION		= EFI_MAX_BIT | 26
EFI_CRC_ERROR				= EFI_MAX_BIT | 27
EFI_END_OF_MEDIA			= EFI_MAX_BIT | 28
EFI_END_OF_FILE				= EFI_MAX_BIT | 31
EFI_INVALID_LANGUAGE		= EFI_MAX_BIT | 32

EFI_WARN_UNKNOWN_GLYPH		= EFI_MAX_BIT | 1
EFI_WARN_DELETE_FAILURE		= EFI_MAX_BIT | 2
EFI_WARN_WRITE_FAILURE		= EFI_MAX_BIT | 3
EFI_WARN_BUFFER_TOO_SMALL	= EFI_MAX_BIT | 4

# @see: MdePkg\Include\Base.h
EFI_ERROR = lambda status: (status & EFI_MAX_BIT) != 0



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ctypes, enum

from qiling.os.macos.structs import POINTER64


base_event_normal = 0
class AutoNumberNormalEvent(enum.Enum):
     def __new__(cls):
        value = len(cls.__members__) + base_event_normal
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

class MacOSEventType(AutoNumberNormalEvent):
    EV_SYSCTL = ()
    EV_PROCESS = ()
    EV_NETWORK = ()
    EV_CTL_CONNECT = ()
    EV_CTL_DISCONNECT = ()
    EV_CTL_SEND = ()
    EV_CTL_SETOPT = ()
    EV_CTL_GETOPT = ()
    EV_CTL_RCVD_FUNC = ()
    EV_CTL_SEND_LIST_FUNC = ()
    EV_CTL_BIND_FUNC = ()
    EV_THREAD = ()
    EV_SFLT_UNREGISTERED = ()
    EV_SFLT_ATTACH = ()
    EV_SFLT_DETACH = ()
    EV_SFLT_NOTIFY_CONNECTING = ()
    EV_SFLT_NOTIFY_CONNECTED = ()
    EV_SFLT_NOTIFY_DISCONNECTING = ()
    EV_SFLT_NOTIFY_DISCONNECTED = ()
    EV_SFLT_NOTIFY_FLUSH_READ = ()
    EV_SFLT_NOTIFY_SHUTDOWN = ()
    EV_SFLT_NOTIFY_CANTRECVMORE = ()
    EV_SFLT_NOTIFY_CANTSENDMORE = ()
    EV_SFLT_NOTIFY_CLOSING = ()
    EV_SFLT_NOTIFY_BOUND = ()
    EV_SFLT_GETPEERNAME = ()
    EV_SFLT_GETSOCKNAME = ()
    EV_SFLT_DATA_IN = ()
    EV_SFLT_DATA_OUT = ()
    EV_SFLT_CONNECT_IN = ()
    EV_SFLT_CONNECT_OUT = ()
    EV_SFLT_BIND = ()
    EV_SFLT_SETOPTION = ()
    EV_SFLT_GETOPTION = ()
    EV_SFLT_LISTEN = ()
    EV_SFLT_IOCTL = ()
    EV_KAUTH_GENERIC = ()
    EV_KAUTH_PROCESS = ()
    EV_KAUTH_VNODE = ()
    EV_KAUTH_FILEOP = ()
    EV_IPF_INPUT = ()
    EV_IPF_OUTPUT = ()
    EV_IPF_DETACH = ()

# enum {
# 	sock_evt_connecting             = 1,
# 	sock_evt_connected              = 2,
# 	sock_evt_disconnecting          = 3,
# 	sock_evt_disconnected           = 4,
# 	sock_evt_flush_read             = 5,
# 	sock_evt_shutdown               = 6, /* param points to an integer specifying how (read, write, or both) see man 2 shutdown */
# 	sock_evt_cantrecvmore           = 7,
# 	sock_evt_cantsendmore           = 8,
# 	sock_evt_closing                = 9,
# 	sock_evt_bound                  = 10
# };

base_event_socket = 0x1000

class SocketEvent(enum.Enum):
    CONNECTING = 0x1001
    CONNECTED = 0x1002
    DISCONNECTING = 0x1003
    DISCONNECTED = 0x1004
    FLUSH_READ = 0x1005
    SHUTDOWN = 0x1006
    CANTRECVMORE = 0x1007
    CANTSENDMORE = 0x1008
    CLOSING = 0x1009
    BOUND = 0x100a
    
class NetworkProtocol(enum.Enum):
    IPPROTO_IP              = 0
    IPPROTO_ICMP            = 1
    IPPROTO_IGMP            = 2
    IPPROTO_GGP             = 3
    IPPROTO_IPV4            = 4
    IPPROTO_TCP             = 6
    IPPROTO_ST              = 7
    IPPROTO_EGP             = 8
    IPPROTO_PIGP            = 9
    IPPROTO_RCCMON          = 10
    IPPROTO_NVPII           = 11
    IPPROTO_PUP             = 12
    IPPROTO_ARGUS           = 13
    IPPROTO_EMCON           = 14
    IPPROTO_XNET            = 15
    IPPROTO_CHAOS           = 16
    IPPROTO_UDP             = 17
    IPPROTO_MUX             = 18
    IPPROTO_MEAS            = 19
    IPPROTO_HMP             = 20
    IPPROTO_PRM             = 21
    IPPROTO_IDP             = 22
    IPPROTO_TRUNK1          = 23
    IPPROTO_TRUNK2          = 24
    IPPROTO_LEAF1           = 25
    IPPROTO_LEAF2           = 26
    IPPROTO_RDP             = 27
    IPPROTO_IRTP            = 28
    IPPROTO_TP              = 29
    IPPROTO_BLT             = 30
    IPPROTO_NSP             = 31
    IPPROTO_INP             = 32
    IPPROTO_SEP             = 33
    IPPROTO_3PC             = 34
    IPPROTO_IDPR            = 35
    IPPROTO_XTP             = 36
    IPPROTO_DDP             = 37
    IPPROTO_CMTP            = 38
    IPPROTO_TPXX            = 39
    IPPROTO_IL              = 40
    IPPROTO_IPV6            = 41
    IPPROTO_SDRP            = 42
    IPPROTO_ROUTING 	    = 43
    IPPROTO_FRAGMENT        = 44
    IPPROTO_IDRP            = 45
    IPPROTO_RSVP            = 46
    IPPROTO_GRE             = 47
    IPPROTO_MHRP            = 48
    IPPROTO_BHA             = 49
    IPPROTO_ESP             = 50
    IPPROTO_AH              = 51
    IPPROTO_INLSP           = 52
    IPPROTO_SWIPE           = 53
    IPPROTO_NHRP            = 54
    IPPROTO_ICMPV6          = 58
    IPPROTO_NONE            = 59
    IPPROTO_DSTOPTS         = 60
    IPPROTO_AHIP            = 61
    IPPROTO_CFTP            = 62
    IPPROTO_HELLO           = 63
    IPPROTO_SATEXPAK        = 64
    IPPROTO_KRYPTOLAN       = 65
    IPPROTO_RVD             = 66
    IPPROTO_IPPC            = 67
    IPPROTO_ADFS            = 68
    IPPROTO_SATMON          = 69
    IPPROTO_VISA            = 70
    IPPROTO_IPCV            = 71
    IPPROTO_CPNX            = 72
    IPPROTO_CPHB            = 73
    IPPROTO_WSN             = 74
    IPPROTO_PVP             = 75
    IPPROTO_BRSATMON        = 76
    IPPROTO_ND              = 77
    IPPROTO_WBMON           = 78
    IPPROTO_WBEXPAK         = 79
    IPPROTO_EON             = 80
    IPPROTO_VMTP            = 81
    IPPROTO_SVMTP           = 82
    IPPROTO_VINES           = 83
    IPPROTO_TTP             = 84
    IPPROTO_IGP             = 85
    IPPROTO_DGP             = 86
    IPPROTO_TCF             = 87
    IPPROTO_IGRP            = 88
    IPPROTO_OSPFIGP         = 89
    IPPROTO_SRPC            = 90
    IPPROTO_LARP            = 91
    IPPROTO_MTP             = 92
    IPPROTO_AX25            = 93
    IPPROTO_IPEIP           = 94
    IPPROTO_MICP            = 95
    IPPROTO_SCCSP           = 96
    IPPROTO_ETHERIP         = 97
    IPPROTO_ENCAP           = 98
    IPPROTO_APES            = 99
    IPPROTO_GMTP            = 100
    IPPROTO_PIM             = 103
    IPPROTO_IPCOMP          = 108
    IPPROTO_PGM             = 113
    IPPROTO_SCTP            = 132
    IPPROTO_DIVERT          = 254
    IPPROTO_RAW             = 255
    IPPROTO_MAX             = 256
    IPPROTO_DONE            = 257

# KAUTH_FILEOP_OPEN                       1
# KAUTH_FILEOP_CLOSE                      2
# KAUTH_FILEOP_RENAME                     3
# KAUTH_FILEOP_EXCHANGE                   4
# KAUTH_FILEOP_LINK                       5
# KAUTH_FILEOP_EXEC                       6
# KAUTH_FILEOP_DELETE                     7
# KAUTH_FILEOP_WILL_RENAME                8
class Kauth(enum.Enum):
    KAUTH_FILEOP_OPEN = 1
    KAUTH_FILEOP_CLOSE = 2
    KAUTH_FILEOP_RENAME = 3
    KAUTH_FILEOP_EXCHANGE = 4
    KAUTH_FILEOP_LINK = 5
    KAUTH_FILEOP_EXEC = 6
    KAUTH_FILEOP_DELETE = 7
    KAUTH_FILEOP_WILL_RENAME = 8

# struct sysctl_oid {
# 	struct sysctl_oid_list *oid_parent;
# 	SLIST_ENTRY(sysctl_oid) oid_link;
# 	int		oid_number;
# 	int		oid_kind;
# 	void		*oid_arg1;
# 	int		oid_arg2;
# 	const char	*oid_name;
# 	int 		(*oid_handler) SYSCTL_HANDLER_ARGS;
# 	const char	*oid_fmt;
# 	const char	*oid_descr; /* offsetof() field / long description */
# 	int		oid_version;
# 	int		oid_refcnt;
# };

class sysctl_oid_t(ctypes.Structure):
    class slist_entry(ctypes.Structure):
        _fields_ = (
            ("sle_next", POINTER64),
        )
    _fields_ = (
        ("oid_parent", POINTER64),
        ("oid_link", slist_entry),
        ("oid_number", ctypes.c_int32),
        ("oid_kind", ctypes.c_int32),
        ("oid_arg1", POINTER64),
        ("oid_arg2", ctypes.c_int32),
        ("oid_name", POINTER64),
        ("oid_handler", POINTER64),
        ("oid_fmt", POINTER64),
        ("oid_descr", POINTER64),
        ("oid_version", ctypes.c_int32),
        ("oid_refcnt", ctypes.c_int32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        for field in self._fields_:
            if isinstance(getattr(self, field[0]), POINTER64):
                print("%s: 0x%x" % (field[0], getattr(self, field[0]).value))
            elif isinstance(getattr(self, field[0]), sysctl_oid_t.slist_entry):
                print("%s: Struct( 0x%x )" % (field[0], getattr(self, field[0]).sle_next.value))
            else:
                print("%s: 0x%x" % (field[0], getattr(self, field[0])))

class sysctl_args_t(ctypes.Structure):
    _fields_ = (
        ("name", ctypes.c_int32 * 2),
	("namelen", ctypes.c_uint32),
	("old", POINTER64),
	("oldlenp", POINTER64),
	("new", POINTER64),
	("newlen", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct sysctlbyname_args {
#     const char * 	name
#     size_t 	namelen
#     void * 	old
#     size_t * 	oldlenp
#     void * 	new
#     size_t 	newlen
#  }

class sysctlbyname_args_t(ctypes.Structure):
    _fields_ = (
        ("name", POINTER64),
        ("namelen", ctypes.c_size_t),
        ("old", POINTER64),
        ("oldlenp", POINTER64),
        ("new", POINTER64),
        ("newlen", ctypes.c_size_t),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct sysctl_req {
# 	struct proc	*p;
# 	int		lock;
# 	user_addr_t	oldptr;		/* pointer to user supplied buffer */
# 	size_t		oldlen;		/* user buffer length (also returned) */
# 	size_t		oldidx;		/* total data iteratively copied out */
# 	int		(*oldfunc)(struct sysctl_req *, const void *, size_t);
# 	user_addr_t	newptr;		/* buffer containing new value */
# 	size_t		newlen;		/* length of new value */
# 	size_t		newidx;		/* total data iteratively copied in */
# 	int		(*newfunc)(struct sysctl_req *, void *, size_t);
# };

class sysctl_req_t(ctypes.Structure):
    _fields_ = (
        ("p", POINTER64),
        ("lock", ctypes.c_int32),
        ("oldptr", POINTER64),
        ("oldlen", ctypes.c_size_t),
        ("oldidx", ctypes.c_size_t),
        ("oldfunc", POINTER64),
        ("newptr", POINTER64),
        ("newlen", ctypes.c_size_t),
        ("newidx", ctypes.c_size_t),
        ("newfunc", ctypes.c_int32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct kern_ctl_reg
# {
# 	/* control information */
# 	char		ctl_name[MAX_KCTL_NAME];
# 	u_int32_t	ctl_id;
# 	u_int32_t	ctl_unit;
# 
#     /* control settings */
#     u_int32_t	ctl_flags;
#     u_int32_t	ctl_sendsize;
#     u_int32_t	ctl_recvsize;
# 
#     /* Dispatch functions */
#     ctl_connect_func	ctl_connect;
#     ctl_disconnect_func	ctl_disconnect;
#     ctl_send_func		ctl_send;
#     ctl_setopt_func		ctl_setopt;
#     ctl_getopt_func		ctl_getopt;
# #ifdef KERNEL_PRIVATE
#     ctl_rcvd_func		ctl_rcvd;	/* Only valid if CTL_FLAG_REG_EXTENDED is set */
#     ctl_send_list_func		ctl_send_list;	/* Only valid if CTL_FLAG_REG_EXTENDED is set */
# 	ctl_bind_func		ctl_bind;
# #endif /* KERNEL_PRIVATE */
# };

class kern_ctl_reg_t(ctypes.Structure):
    _fields_ = (
        ("ctl_name", ctypes.c_char * 96),
        ("ctl_id", ctypes.c_uint32),
        ("ctl_unit", ctypes.c_uint32),
        ("ctl_flags", ctypes.c_uint32),
        ("ctl_sendsize", ctypes.c_uint32),
        ("ctl_recvsize", ctypes.c_uint32),
        ("ctl_connect", POINTER64),
        ("ctl_disconnect", POINTER64),
        ("ctl_send", POINTER64),
        ("ctl_setopt", POINTER64),
        ("ctl_getopt", POINTER64),
        ("ctl_rcvd", POINTER64),
        ("ctl_send_list", POINTER64),
        ("ctl_bind", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        for field in self._fields_:
            if isinstance(getattr(self, field[0]), POINTER64):
                print("%s: 0x%x" % (field[0], getattr(self, field[0]).value))
            elif isinstance(getattr(self, field[0]), int):
                print("%s: %d" % (field[0], getattr(self, field[0])))
            elif isinstance(getattr(self, field[0]), bytes):
                print("%s: %s" % (field[0], getattr(self, field[0]).decode()))


# struct sockaddr_ctl {
#     u_char	sc_len;		/* depends on size of bundle ID string */
#     u_char	sc_family;	/* AF_SYSTEM */
#     u_int16_t 	ss_sysaddr;	/* AF_SYS_KERNCONTROL */
#     u_int32_t	sc_id; 		/* Controller unique identifier  */
#     u_int32_t 	sc_unit;	/* Developer private unit number */
#     u_int32_t 	sc_reserved[5];
# };

class sockaddr_ctl_t(ctypes.Structure):
    _fields_ = (
        ("sc_len", ctypes.c_ubyte),
        ("sc_family", ctypes.c_ubyte),
        ("ss_sysaddr", ctypes.c_uint16),
        ("sc_id", ctypes.c_uint32),
        ("sc_unit", ctypes.c_uint32),
        ("sc_reserved", ctypes.c_uint32 * 5),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct m_hdr {
# 	struct mbuf	*mh_next;	/* next buffer in chain */
# 	struct mbuf	*mh_nextpkt;	/* next chain in queue/record */
# 	caddr_t		mh_data;	/* location of data */
# 	int32_t		mh_len;		/* amount of data in this mbuf */
# 	u_int16_t	mh_type;	/* type of data in this mbuf */
# 	u_int16_t	mh_flags;	/* flags; see below */
# }

class m_hdr_t(ctypes.Structure):
    _fields_ = (
        ("mh_next", POINTER64),
        ("mh_nextpkt", POINTER64),
        ("mh_data", POINTER64),
        ("mh_len", ctypes.c_int32),
        ("mh_type", ctypes.c_uint16),
        ("mh_flags", ctypes.c_uint16),
    )

class tag_t(ctypes.Structure):
    _fields_ = (
        ("packet_tags", POINTER64),
    )

# struct tcp_pktinfo {
#     union {
#         struct {
#                 u_int32_t segsz;	/* segment size (actual MSS) */
#                 u_int32_t start_seq;	/* start seq of this packet */
#         } __tx;
#         struct {
#                 u_int16_t lro_pktlen;	/* max seg size encountered */
#                 u_int8_t  lro_npkts;	/* # of coalesced TCP pkts */
#                 u_int8_t  lro_timediff;	/* time spent in LRO */
#         } __rx;
#     } __offload;
#     union {
#         u_int32_t	pri;		/* send msg priority */
#         u_int32_t	seq;		/* recv msg sequence # */
#     } __msgattr;
# };
class tcp_pktinfo_t(ctypes.Structure):
    class __offload_u(ctypes.Union):
        class __tx_t(ctypes.Structure):
            _fields_ = (
                ("segsz", ctypes.c_uint32),
                ("start_seq", ctypes.c_uint32),
            )
        class __rx_t(ctypes.Structure):
            _fields_ = (
                ("lro_pktlen", ctypes.c_uint16),
                ("lro_npkts", ctypes.c_uint8),
                ("lro_timediff", ctypes.c_uint8),
            )
        _fields_ = (
            ("__tx", __tx_t),
            ("__rx", __rx_t),
        )
    class __msgattr_u(ctypes.Union):
        _fields_ = (
            ("pri", ctypes.c_uint32),
            ("seq", ctypes.c_uint32),
        )
    _fields_ = (
        ("__offload", __offload_u),
        ("__msgattr", __msgattr_u),
    )

# struct mptcp_pktinfo {
#     u_int64_t	mtpi_dsn;	/* MPTCP Data Sequence Number */
#     u_int32_t	mtpi_rel_seq;	/* Relative Seq Number */
#     u_int16_t	mtpi_length;	/* Length of mapping */
#     u_int16_t	mtpi_csum;
# };
class mptcp_pktinfo_t(ctypes.Structure):
    _fields_ = (
        ("mtpi_dsn", ctypes.c_uint64),
        ("mtpi_rel_seq", ctypes.c_uint32),
        ("mtpi_length", ctypes.c_uint16),
        ("mtpi_csum", ctypes.c_uint16),
    )

# struct tcp_mtag {
#     union {
#         struct tcp_pktinfo	tm_tcp;		/* TCP and below */
#         struct mptcp_pktinfo	tm_mptcp;	/* MPTCP-TCP only */
#     };
# };
class tcp_mtag_t(ctypes.Structure):
    class pktinfo_u(ctypes.Union):
        _fields_ = (
            ("tm_tcp", tcp_pktinfo_t),
            ("tm_mptcp", mptcp_pktinfo_t),
        )
    _anonymous_ = ("tmp_union",)
    _fields_ = (
        ("tmp_union", pktinfo_u),
    )

# struct proto_mtag_ {
#     union {
#         struct tcp_mtag	tcp;		/* TCP specific */
#     } __pr_u;
# };
class proto_mtag__t(ctypes.Structure):
    class __pr_u_u(ctypes.Union):
        _fields_ = (
            ("tcp", tcp_mtag_t),
        )
    _fields_ = (
        ("__pr_u", __pr_u_u),
    )

# struct pf_mtag {
#     u_int16_t	pftag_flags;	/* PF_TAG flags */
#     u_int16_t	pftag_rtableid;	/* alternate routing table id */
#     u_int16_t	pftag_tag;
#     u_int16_t	pftag_routed;
# #if PF_ECN
#     void		*pftag_hdr;	/* saved hdr pos in mbuf, for ECN */
# #endif /* PF_ECN */
# };
class pf_mtag_t(ctypes.Structure):
    _fields_ = (
        ("pftag_flags", ctypes.c_int16),
        ("pftag_rtableid", ctypes.c_int16),
        ("pftag_tag", ctypes.c_int16),
        ("pftag_routed", ctypes.c_int16),
    )

# struct necp_mtag_ {
#     u_int32_t	necp_policy_id;
#     u_int32_t	necp_last_interface_index;
#     u_int32_t	necp_route_rule_id;
#     u_int32_t	necp_app_id;
# };
class necp_mtag__t(ctypes.Structure):
    _fields_ = (
        ("necp_policy_id", ctypes.c_int32),
        ("necp_last_interface_index", ctypes.c_int32),
        ("necp_route_rule_id", ctypes.c_int32),
        ("necp_app_id", ctypes.c_int32),
    )



# struct {
#     union {
#         u_int8_t	__mpriv8[16];
#         u_int16_t	__mpriv16[8];
#         struct {
#             union {
#                 u_int8_t	__val8[4];
#                 u_int16_t	__val16[2];
#                 u_int32_t	__val32;
#             } __mpriv32_u;
#         } __mpriv32[4];
#         u_int64_t	__mpriv64[2];
#     } __mpriv_u;
# } pkt_mpriv __attribute__((aligned(4)));
class pkt_mpriv_t(ctypes.Structure):
    class __mpriv_u_u(ctypes.Union):
        class __mpriv32_t(ctypes.Structure):
            class __mpriv32_u_u(ctypes.Union):
                _fields_ = (
                    ("__val8", ctypes.c_int8 * 4),
                    ("__val16", ctypes.c_int16 * 2),
                    ("__val32", ctypes.c_int32),
                )
            _fields_ = (
                ("__mpriv32_u", __mpriv32_u_u),
            )
        _fields_ = (
            ("__mpriv32", __mpriv32_t * 4),
            ("__mpriv64", ctypes.c_uint64 * 2),
        )
    _fields_ = (
        ("__mpriv_u", __mpriv_u_u),
    )

# struct pkthdr {
# 	struct ifnet *rcvif;		/* rcv interface */
# 	void	*pkt_hdr;		/* pointer to packet header */
# 	int32_t	len;			/* total packet length */
# 	u_int32_t csum_flags;		/* flags regarding checksum */
# 	union {
# 		struct {
# 			u_int16_t val;	 /* checksum value */
# 			u_int16_t start; /* checksum start offset */
# 		} _csum_rx;
# 		struct {
# 			u_int16_t start; /* checksum start offset */
# 			u_int16_t stuff; /* checksum stuff offset */
# 		} _csum_tx;
# 		u_int32_t csum_data;	/* data field used by csum routines */
# 	};
# 	u_int16_t vlan_tag;		/* VLAN tag, host byte order */
# 	u_int8_t pkt_proto;		/* IPPROTO value */
# 	u_int8_t pkt_flowsrc;		/* FLOWSRC values */
# 	u_int32_t pkt_flowid;		/* flow ID */
# 	u_int32_t pkt_flags;		/* PKTF flags (see below) */
# 	u_int32_t pkt_svc;		/* MBUF_SVC value */
# 
# 	u_int32_t pkt_compl_context;		/* Packet completion context */
# 
# 	union {
# 		struct {
# 			u_int16_t src;		/* ifindex of src addr i/f */
# 			u_int16_t src_flags;	/* src PKT_IFAIFF flags */
# 			u_int16_t dst;		/* ifindex of dst addr i/f */
# 			u_int16_t dst_flags;	/* dst PKT_IFAIFF flags */
# 		} _pkt_iaif;
# 		u_int64_t pkt_ifainfo;	/* data field used by ifainfo */
# 		struct {
# 			u_int32_t if_data; /* bytes in interface queue */
# 			u_int32_t sndbuf_data; /* bytes in socket buffer */
# 		} _pkt_bsr;	/* Buffer status report used by cellular interface */
# 	};
# 	u_int64_t pkt_timestamp;	/* enqueue time */
# 
# 	SLIST_HEAD(packet_tags, m_tag) tags; /* list of external tags */
# 	union builtin_mtag builtin_mtag;
# 	struct {
# 		union {
# 			u_int8_t	__mpriv8[16];
# 			u_int16_t	__mpriv16[8];
# 			struct {
# 				union {
# 					u_int8_t	__val8[4];
# 					u_int16_t	__val16[2];
# 					u_int32_t	__val32;
# 				} __mpriv32_u;
# 			}		__mpriv32[4];
# 			u_int64_t	__mpriv64[2];
# 		} __mpriv_u;
# 	} pkt_mpriv __attribute__((aligned(4)));
# 	u_int32_t redzone;		/* red zone */
# 	u_int32_t pkt_compl_callbacks;	/* Packet completion callbacks */
# };

class pkthdr_t(ctypes.Structure):
    class chksum_union(ctypes.Union):
        class _csum_rx_t(ctypes.Structure):
            _fields_ = (
                ("val", ctypes.c_uint16),
                ("start", ctypes.c_uint16),
            )
        class _csum_tx_t(ctypes.Structure):
            _fields_ = (
                ("start", ctypes.c_uint16),
                ("stuff", ctypes.c_uint16),
            )
        _fields_ = [
            ("_csum_rx", _csum_rx_t),
            ("_csum_tx", _csum_tx_t),
            ("csum_data", ctypes.c_uint32),
        ]

    class interface_union(ctypes.Union):
        class _pkt_iaif_t(ctypes.Structure):
            _fields_ = (
                ("src", ctypes.c_uint16),
                ("src_flags", ctypes.c_uint16),
                ("dst", ctypes.c_uint16),
                ("dst_flags", ctypes.c_uint16),
            )
        class _pkt_bsr_t(ctypes.Structure):
            _fields_ = (
                ("if_data", ctypes.c_uint32),
                ("sndbuf_data", ctypes.c_uint32),
            )
        _fields_ = (
            ("_pkt_iaif", _pkt_iaif_t),
            ("pkt_ifainfo", ctypes.c_uint64),
            ("_pkt_bsr", _pkt_bsr_t),
        )

#     union builtin_mtag {
# 	struct {
# 		struct proto_mtag_ _proto_mtag;	/* built-in protocol-specific tag */
# 		struct pf_mtag	_pf_mtag;	/* built-in PF tag */
# 		struct necp_mtag_ _necp_mtag; /* built-in NECP tag */
# 	} _net_mtag;
# 	struct driver_mtag_ _drv_mtag;
#     }
    class builtin_mtag_u(ctypes.Union):
#         struct driver_mtag_ {
#             uintptr_t		_drv_tx_compl_arg;
#             uintptr_t		_drv_tx_compl_data;
#             kern_return_t		_drv_tx_status;
#             uint16_t		_drv_flowid;
#         };
        class driver_mtag__t(ctypes.Structure):
            _fields_ = (
                ("_drv_tx_compl_arg", POINTER64),
                ("_drv_tx_compl_data", POINTER64),
                ("_drv_tx_status", ctypes.c_int32),
                ("_drv_flowid", ctypes.c_int16),
            )
        class _net_mtag_t(ctypes.Structure):
            _fields_ = (
                ("_proto_mtag", proto_mtag__t),
                ("_pf_mtag", pf_mtag_t),
                ("_necp_mtag", necp_mtag__t),
            )
        _fields_ = (
            ("_net_mtag", _net_mtag_t),
            ("_drv_mtag", driver_mtag__t),
        )

    _anonymous_ = ("tmp_chksum_union", "tmp_interface_union", )
    _fields_ = (
        ("rcvif", POINTER64),
        ("pkt_hdr", POINTER64),
        ("len", ctypes.c_int32),
        ("csum_flags", ctypes.c_uint32),
        ("tmp_chksum_union", chksum_union),
        ("vlan_tag", ctypes.c_uint16),
        ("pkt_proto", ctypes.c_uint8),
        ("pkt_flowsrc", ctypes.c_uint8),
        ("pkt_flowid", ctypes.c_uint32),
        ("pkt_flags", ctypes.c_uint32),
        ("pkt_svc", ctypes.c_uint32),
        ("pkt_compl_context", ctypes.c_uint32),
        ("tmp_interface_union", interface_union),
        ("pkt_timestamp", ctypes.c_uint64),
        ("tags", tag_t),
        ("builtin_mtag", builtin_mtag_u),
        ("pkt_mpriv", pkt_mpriv_t),
        ("redzone", ctypes.c_uint32),
        ("pkt_compl_callbacks", ctypes.c_uint32),
    )

# struct m_ext {
# 	caddr_t	ext_buf;		/* start of buffer */
# 	m_ext_free_func_t ext_free;	/* free routine if not the usual */
# 	u_int	ext_size;		/* size of buffer, for ext_free */
# 	caddr_t	ext_arg;		/* additional ext_free argument */
# 	struct ext_ref {
# 		struct mbuf *paired;
# 		u_int16_t minref;
# 		u_int16_t refcnt;
# 		u_int16_t prefcnt;
# 		u_int16_t flags;
# 		u_int32_t priv;
# 		uintptr_t ext_token;
# 	} *ext_refflags;
# };
class ext_ref(ctypes.Structure):
    _fields_ = (
        ("paired", POINTER64),
        ("minref", ctypes.c_uint16),
        ("refcnt", ctypes.c_uint16),
        ("prefcnt", ctypes.c_uint16),
        ("flags", ctypes.c_uint16),
        ("priv", ctypes.c_uint32),
        ("ext_token", POINTER64),
    )
class m_ext_t(ctypes.Structure):
    _fields_ = (
        ("ext_buf", POINTER64),
        ("ext_free", POINTER64),
        ("ext_size", ctypes.c_uint32),
        ("ext_arg", POINTER64),
        ("ext_refflags", POINTER64),
    )

# struct mbuf {
# 	struct m_hdr m_hdr;
# 	union {
# 		struct {
# 			struct pkthdr MH_pkthdr;	/* M_PKTHDR set */
# 			union {
# 				struct m_ext MH_ext;	/* M_EXT set */
# 				char	MH_databuf[_MHLEN];
# 			} MH_dat;
# 		} MH;
# 		char	M_databuf[_MLEN];		/* !M_PKTHDR, !M_EXT */
# 	} M_dat;
# };

#define	MSIZESHIFT	8			/* 256 */
#define	MSIZE		(1 << MSIZESHIFT)	/* size of an mbuf */
#define	_MLEN		(MSIZE - sizeof(struct m_hdr))	/* normal data len */
#define	_MHLEN		(_MLEN - sizeof(struct pkthdr))	/* data len w/pkthdr */

MSIZESHIFT = 8
MSIZE = (1 << MSIZESHIFT)
_MLEN = (MSIZE - ctypes.sizeof(m_hdr_t))
_MHLEN = (_MLEN - ctypes.sizeof(pkthdr_t))

class mbuf_t(ctypes.Structure):
    class M_dat_u(ctypes.Union):
        class MH_t(ctypes.Structure):
            class MH_dat_u(ctypes.Union):
                _fields_ = (
                    ("MH_ext", m_ext_t),
                    ("MH_databuf", ctypes.c_char * _MHLEN)
                )
            _fields_ = (
                ("MH_pkthdr", pkthdr_t),
                ("MH_dat", MH_dat_u),
            )
        _fields_ = (
            ("MH", MH_t),
            ("M_databuf", ctypes.c_char * _MLEN),
        )
    _fields_ = (
        ("m_hdr", m_hdr_t),
        ("M_dat", M_dat_u),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# enum sopt_dir { SOPT_GET, SOPT_SET };
# struct sockopt {
# 	enum sopt_dir sopt_dir; /* is this a get or a set? */
# 	int	sopt_level;	/* second arg of [gs]etsockopt */
# 	int	sopt_name;	/* third arg of [gs]etsockopt */
# 	void* sopt_val;	/* fourth arg of [gs]etsockopt */
# 	size_t	sopt_valsize;	/* (almost) fifth arg of [gs]etsockopt */
# 	void *sopt_p;	/* calling process or null if kernel */
# };

class sockopt_t(ctypes.Structure):
    _fields_ = (
        ("sopt_dir", ctypes.c_uint64),
        ("sopt_level", ctypes.c_int32),
        ("sopt_name", ctypes.c_int32),
        ("sopt_val", POINTER64),
        ("sopt_valsize", ctypes.c_uint64),
        ("sopt_p", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct sflt_filter {
# 	sflt_handle                     sf_handle;
# 	int                             sf_flags;
# 	char                            *sf_name;
# 
# 	sf_unregistered_func            sf_unregistered;
# 	sf_attach_func                  sf_attach;
# 	sf_detach_func                  sf_detach;
# 
# 	sf_notify_func                  sf_notify;
# 	sf_getpeername_func             sf_getpeername;
# 	sf_getsockname_func             sf_getsockname;
# 	sf_data_in_func                 sf_data_in;
# 	sf_data_out_func                sf_data_out;
# 	sf_connect_in_func              sf_connect_in;
# 	sf_connect_out_func             sf_connect_out;
# 	sf_bind_func                    sf_bind;
# 	sf_setoption_func               sf_setoption;
# 	sf_getoption_func               sf_getoption;
# 	sf_listen_func                  sf_listen;
# 	sf_ioctl_func                   sf_ioctl;
# 	/*
# 	 * The following are valid only if SFLT_EXTENDED flag is set.
# 	 * Initialize sf_ext_len to sizeof sflt_filter_ext structure.
# 	 * Filters must also initialize reserved fields with zeroes.
# 	 */
# 	struct sflt_filter_ext {
# 		unsigned int            sf_ext_len;
# 		sf_accept_func          sf_ext_accept;
# 		void                    *sf_ext_rsvd[5];        /* Reserved */
# 	} sf_ext;
# };

class sflt_filter_t(ctypes.Structure):
    class sflt_filter_ext(ctypes.Structure):
        _fields_ = (
            ("sf_ext_len", ctypes.c_uint32),
            ("sf_ext_accept", POINTER64),
            ("sf_ext_rsvd", POINTER64 * 5),
        )

    _fields_ = (
        ("sf_handle", ctypes.c_uint32),
        ("sf_flags", ctypes.c_int32),
        ("sf_name", POINTER64),
        ("sf_unregistered", POINTER64),
        ("sf_attach", POINTER64),
        ("sf_detach", POINTER64),
        ("sf_notify", POINTER64),
        ("sf_getpeername", POINTER64),
        ("sf_getsockname", POINTER64),
        ("sf_data_in", POINTER64),
        ("sf_data_out", POINTER64),
        ("sf_connect_in", POINTER64),
        ("sf_connect_out", POINTER64),
        ("sf_bind", POINTER64),
        ("sf_setoption", POINTER64),
        ("sf_getoption", POINTER64),
        ("sf_listen", POINTER64),
        ("sf_ioctl", POINTER64),
        ("sf_ext", sflt_filter_ext),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        self.ql.log.info("[*] Dumping object: %s" % (self.sf_name))
        for field in self._fields_:
            if isinstance(getattr(self, field[0]), POINTER64):
                self.ql.log.info("%s: 0x%x" % (field[0], getattr(self, field[0]).value))
            elif isinstance(getattr(self, field[0]), int):
                self.ql.log.info("%s: %d" % (field[0], getattr(self, field[0])))
            elif isinstance(getattr(self, field[0]), bytes):
                self.ql.log.info("%s: %s" % (field[0], getattr(self, field[0]).decode()))

# struct sockaddr_in {
# 	__uint8_t	sin_len;
# 	sa_family_t	sin_family;
# 	in_port_t	sin_port;
# 	struct	in_addr sin_addr;
# 	char		sin_zero[8];
# };

class sockaddr_in_t(ctypes.Structure):
#     struct in_addr {
# 	in_addr_t s_addr;
#     };
    class in_addr_t(ctypes.Structure):
        _fields_ = (
            ("s_addr", ctypes.c_uint32),
        )
    _fields_ = (
        ("sin_len", ctypes.c_uint8),
        ("sin_family", ctypes.c_uint8),
        ("sin_port", ctypes.c_uint16),
        ("sin_addr", in_addr_t),
        ("sin_zero", ctypes.c_char * 8),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# #define ETHER_ADDR_LEN          6
# typedef struct  ether_header {
# 	u_char  ether_dhost[ETHER_ADDR_LEN];
# 	u_char  ether_shost[ETHER_ADDR_LEN];
# 	u_short ether_type;
# } ether_header_t;
class ether_header_t(ctypes.Structure):
    _fields_ = (
        ("ether_dhost", ctypes.c_ubyte * 6),
        ("ether_shost", ctypes.c_ubyte * 6),
        ("ether_type", ctypes.c_ushort),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct mac_policy_conf {
#     char *mpc_name;                    // policy name
#     char *mpc_fullname;                // full name
#     char const *const *mpc_labelnames; // managed label namespaces
#     unsigned int mpc_labelname_count;  // number of managed label namespaces
#     struct mac_policy_ops *mpc_ops;    // operation vector
#     int mpc_loadtime_flags;            // load time flags
#     int *mpc_field_off;                // label slot
#     int mpc_runtime_flags;             // run time flags
#     struct mac_policy_conf *mpc_list;  // list reference
#     void *mpc_data;                    // module data
# };
class mac_policy_conf_t(ctypes.Structure):
    _fields_ = (
        ("mpc_name", POINTER64),
        ("mpc_fullname", POINTER64),
        ("mpc_labelnames", POINTER64),
        ("mpc_labelname_count", ctypes.c_uint32),
        ("mpc_ops", POINTER64),
        ("mpc_loadtime_flags", ctypes.c_int32),
        ("mpc_field_off", POINTER64),
        ("mpc_runtime_flags", ctypes.c_int32),
        ("mpc_list", POINTER64),
        ("mpc_data", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        for field in self._fields_:
            if isinstance(getattr(self, field[0]), POINTER64):
                self.ql.log.info("%s: 0x%x" % (field[0], getattr(self, field[0]).value))
            elif isinstance(getattr(self, field[0]), int):
                self.ql.log.info("%s: %d" % (field[0], getattr(self, field[0])))
            elif isinstance(getattr(self, field[0]), bytes):
                self.ql.log.info("%s: %s" % (field[0], getattr(self, field[0]).decode()))

# struct ucred {
# 	TAILQ_ENTRY(ucred)	cr_link; /* never modify this without KAUTH_CRED_HASH_LOCK */
# 	u_long	cr_ref;			/* reference count */
# 	
# struct posix_cred {
# 	uid_t	cr_uid;			/* effective user id */
# 	uid_t	cr_ruid;		/* real user id */
# 	uid_t	cr_svuid;		/* saved user id */
# 	short	cr_ngroups;		/* number of groups in advisory list */
# 	gid_t	cr_groups[NGROUPS];	/* advisory group list */
# 	gid_t	cr_rgid;		/* real group id */
# 	gid_t	cr_svgid;		/* saved group id */
# 	uid_t	cr_gmuid;		/* UID for group membership purposes */
# 	int	cr_flags;		/* flags on credential */
# } cr_posix;
# 	struct label	*cr_label;	/* MAC label */
# 	struct au_session cr_audit;		/* user auditing data */
# };

class ucred_t(ctypes.Structure):
    class cr_entry(ctypes.Structure):
        _fields_ = (
            ("tqe_next", POINTER64),
            ("tqe_prev", POINTER64),
        )
    class posix_cred_t(ctypes.Structure):
        _fields_ = (
            ("cr_uid", ctypes.c_uint32),
            ("cr_ruid", ctypes.c_uint32),
            ("cr_svuid",  ctypes.c_uint32),
            ("cr_ngroups", ctypes.c_short),
            ("cr_groups", ctypes.c_uint32 * 16),
            ("cr_rgid", ctypes.c_uint32),
            ("cr_svgid", ctypes.c_uint32),
            ("cr_gmuid", ctypes.c_uint32),
            ("cr_flags", ctypes.c_int32),
        )
    class au_session_t(ctypes.Structure):
        _fields_ = (
            ("as_aia_p", POINTER64),
            ("as_mask", POINTER64),
        )
    _fields_ = (
        ("cr_link", cr_entry),
        ("cr_ref", ctypes.c_ulong),
        ("cr_posix", posix_cred_t),
        ("cr_label", POINTER64),
        ("cr_audit", au_session_t),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct label {
#     int	l_flags;
#     union {
#             void	*l_ptr;
#             long	 l_long;
#     }	l_perpolicy[MAC_MAX_SLOTS];
# };

class label_t(ctypes.Structure):
    class l_perpolicy_t(ctypes.Union):
        _fields_ = (
            ("l_ptr", POINTER64),
            ("l_long", ctypes.c_long),
        )
    _fields_ = (
        ("l_flags", ctypes.c_int32),
        ("l_perpolicy", l_perpolicy_t * 7)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct vnode {
# 	lck_mtx_t v_lock;			/* vnode mutex */
# 	TAILQ_ENTRY(vnode) v_freelist;		/* vnode freelist */
# 	TAILQ_ENTRY(vnode) v_mntvnodes;		/* vnodes for mount point */
#         TAILQ_HEAD(, namecache) v_ncchildren;	/* name cache entries that regard us as their parent */
#         LIST_HEAD(, namecache) v_nclinks;	/* name cache entries that name this vnode */
#         vnode_t	 v_defer_reclaimlist;		/* in case we have to defer the reclaim to avoid recursion */
#         uint32_t v_listflag;			/* flags protected by the vnode_list_lock (see below) */
# 	uint32_t v_flag;			/* vnode flags (see below) */
# 	uint16_t v_lflag;			/* vnode local and named ref flags */
# 	uint8_t	 v_iterblkflags;		/* buf iterator flags */
# 	uint8_t	 v_references;			/* number of times io_count has been granted */
# 	int32_t	 v_kusecount;			/* count of in-kernel refs */
# 	int32_t	 v_usecount;			/* reference count of users */
# 	int32_t	 v_iocount;			/* iocounters */
# 	void *   v_owner;			/* act that owns the vnode */
# 	uint16_t v_type;			/* vnode type */
# 	uint16_t v_tag;				/* type of underlying data */
# 	uint32_t v_id;				/* identity of vnode contents */
# 	union {
# 		struct mount	*vu_mountedhere;/* ptr to mounted vfs (VDIR) */
# 		struct socket	*vu_socket;	/* unix ipc (VSOCK) */
# 		struct specinfo	*vu_specinfo;	/* device (VCHR, VBLK) */
# 		struct fifoinfo	*vu_fifoinfo;	/* fifo (VFIFO) */
# 	        struct ubc_info *vu_ubcinfo;	/* valid for (VREG) */
# 	} v_un;
# 	struct	buflists v_cleanblkhd;		/* clean blocklist head */
# 	struct	buflists v_dirtyblkhd;		/* dirty blocklist head */
# 	struct klist v_knotes;			/* knotes attached to this vnode */
#         kauth_cred_t	v_cred;			/* last authorized credential */
#         kauth_action_t	v_authorized_actions;	/* current authorized actions for v_cred */
#         int		v_cred_timestamp;	/* determine if entry is stale for MNTK_AUTH_OPAQUE */
#         int		v_nc_generation;	/* changes when nodes are removed from the name cache */
# 	int32_t		v_numoutput;			/* num of writes in progress */
# 	int32_t		v_writecount;			/* reference count of writers */
# 	const char *v_name;			/* name component of the vnode */
# 	vnode_t v_parent;			/* pointer to parent vnode */
# 	struct lockf	*v_lockf;		/* advisory lock list head */
# 	int 	(**v_op)(void *);		/* vnode operations vector */
# 	mount_t v_mount;			/* ptr to vfs we are in */
# 	void *	v_data;				/* private data for fs */
# 	struct label *v_label;			/* MAC security label */
# 	vnode_resolve_t v_resolve;		/* trigger vnode resolve info (VDIR only) */
# };
class vnode_t(ctypes.Structure):
    class tailq_entry(ctypes.Structure):
        _fields_ = (
            ("tqe_next", POINTER64),
            ("tqe_prev", POINTER64),
        )
    class tailq_head(ctypes.Structure):
        _fields_ = (
            ("tqh_first", POINTER64),
            ("tqh_last", POINTER64),
        )
    class list_head(ctypes.Structure):
        _fields_ = (
            ("v_nclinks", POINTER64),
        )
    class slist_head(ctypes.Structure):
        _fields_ = (
            ("slh_first", POINTER64),
        )
    class v_un_t(ctypes.Union):
        _fields_ = (
            ("vu_mountedhere", POINTER64),
            ("vu_socket", POINTER64),
            ("vu_specinfo", POINTER64),
            ("vu_fifoinfo", POINTER64),
            ("vu_ubcinfo", POINTER64),
        )
    _fields_ = (
        ("v_lock", POINTER64),
        ("v_freelist", tailq_entry),
        ("v_mntvnodes", tailq_entry),
        ("v_ncchildren", tailq_head),
        ("v_nclinks", list_head),
        ("v_defer_reclaimlist", POINTER64),
        ("v_listflag", ctypes.c_uint32),
        ("v_flag", ctypes.c_uint32),
        ("v_lflag", ctypes.c_uint16),
        ("v_iterblkflags", ctypes.c_uint8),
        ("v_references", ctypes.c_uint8),
        ("v_kusecount", ctypes.c_uint32),
        ("v_usecount", ctypes.c_uint32),
        ("v_iocount", ctypes.c_uint32),
        ("v_owner", POINTER64),
        ("v_type", ctypes.c_uint16),
        ("v_tag", ctypes.c_uint16),
        ("v_id", ctypes.c_uint32),
        ("v_un", v_un_t),
        ("v_cleanblkhd", list_head),
        ("v_dirtyblkhd", list_head),
        ("v_knotes", slist_head),
        ("v_cred", POINTER64),
        ("v_authorized_actions", ctypes.c_int32),
        ("v_cred_timestamp", ctypes.c_int),
        ("v_nc_generation", ctypes.c_int),
        ("v_numoutput", ctypes.c_int32),
        ("v_writecount", ctypes.c_int32),
        ("v_name", POINTER64),
        ("v_parent", POINTER64),
        ("v_lockf", POINTER64),
        ("v_op", POINTER64),
        ("v_mount", POINTER64),
        ("v_data", POINTER64),
        ("v_label", POINTER64),
        ("v_resolve", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct fileglob {
# 	LIST_ENTRY(fileglob) f_msglist;/* list of active files */
# 	int32_t	fg_flag;		/* see fcntl.h */
# 	int32_t	fg_count;	/* reference count */
# 	int32_t	fg_msgcount;	/* references from message queue */
# 	int32_t fg_lflags;	/* file global flags */
# 	kauth_cred_t fg_cred;	/* credentials associated with descriptor */
# 	const struct fileops {
# 		file_type_t	fo_type;	/* descriptor type */
# 		int	(*fo_read)	(struct fileproc *fp, struct uio *uio,
# 					 int flags, vfs_context_t ctx);
# 		int	(*fo_write)	(struct fileproc *fp, struct uio *uio,
# 					 int flags, vfs_context_t ctx);
# 		int	(*fo_ioctl)	(struct fileproc *fp, u_long com,
# 					 caddr_t data, vfs_context_t ctx);
# 		int	(*fo_select)	(struct fileproc *fp, int which,
# 					 void *wql, vfs_context_t ctx);
# 		int	(*fo_close)	(struct fileglob *fg, vfs_context_t ctx);
# 		int	(*fo_kqfilter)	(struct fileproc *fp, struct knote *kn,
# 					 struct kevent_internal_s *kev, vfs_context_t ctx);
# 		int	(*fo_drain)	(struct fileproc *fp, vfs_context_t ctx);
# 	} *fg_ops;
# 	off_t	fg_offset;
# 	void 	*fg_data;	/* vnode or socket or SHM or semaphore */
# 	void	*fg_vn_data;	/* Per fd vnode data, used for directories */
# 	lck_mtx_t fg_lock;
# 	struct label *fg_label;  /* JMM - use the one in the cred? */
# };

class fileglob_t(ctypes.Structure):
    class list_entry(ctypes.Structure):
        _fields_ = (
            ("le_next", POINTER64),
            ("le_prev", POINTER64),
        )
    _fields_ = (
        ("f_msglist", list_entry),
        ("fg_flag", ctypes.c_int32),
        ("fg_count", ctypes.c_int32),
        ("fg_msgcount", ctypes.c_int32),
        ("fg_lflags", ctypes.c_int32),
        ("fg_cred", POINTER64),
        ("fg_ops", POINTER64),
        ("fg_offset", ctypes.c_int64),
        ("fg_data", POINTER64),
        ("fg_vn_data", POINTER64),
        ("fg_lock", POINTER64),
        ("fg_label", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct mac_policy_list_element {
#     struct mac_policy_conf *mpc;
# };
# 
# struct mac_policy_list {
#     u_int numloaded;
#     u_int max;
#     u_int maxindex;
#     u_int staticmax;
#     u_int chunks;
#     u_int freehint;
#     struct mac_policy_list_element *entries;
# };
class mac_policy_list_element_t(ctypes.Structure):
    _fields_ = (
        ("mpc", POINTER64),
    )

class mac_policy_list_t(ctypes.Structure):
    _fields_ = (
        ("numloaded", ctypes.c_uint),
        ("max", ctypes.c_uint),
        ("maxindex", ctypes.c_uint),
        ("staticmax", ctypes.c_uint),
        ("chunks", ctypes.c_uint),
        ("freehint", ctypes.c_uint),
        ("entries", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct ipf_filter {
# 	void		*cookie;
# 	const char	*name;
# 	ipf_input_func	ipf_input;
# 	ipf_output_func	ipf_output;
# 	ipf_detach_func	ipf_detach;
# };
class ipf_filter_t(ctypes.Structure):
    _fields_ = (
        ("cookie", POINTER64),
        ("name", POINTER64),
        ("ipf_input", POINTER64),
        ("ipf_output", POINTER64),
        ("ipf_detach", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import random
from enum import Enum

class CaneryType(Enum):
        underflow = 0
        overflow = 1
        uaf = 2

class QlSanitizedMemoryHeap():
    """
    Clients can enable the santized heap using the following snippet:

    ql.os.heap = qiling.os.memory.QlSanitizedMemoryHeap(ql, ql.os.heap)
    ql.os.heap.oob_handler = my_oob_handler
    ql.os.heap.bo_handler = my_bo_handler
    ql.os.heap.bad_free_handler = my_bad_free_handler
    ql.os.heap.uaf_handler = my_uaf_handler
    """

    def __init__(self, ql, heap, fault_rate=0, canary_byte=b'\xCD'):
        self.ql = ql
        self.heap = heap
        self.fault_rate = fault_rate
        self.canary_byte = canary_byte
        self.canaries = []

    def save(self):
        saved_state = {}
        saved_state['heap'] = self.heap.save()
        saved_state['fault_rate'] = self.fault_rate
        saved_state['canary_byte'] = self.canary_byte
        saved_state['canaries'] = self.canaries
        return saved_state

    def restore(self, saved_state):
        self.heap.restore(saved_state['heap'])
        self.fault_rate = saved_state['fault_rate']
        self.canary_byte = saved_state['canary_byte']
        self.canaries = saved_state['canaries']
        for (canary_begin, canary_end, canery_type) in self.canaries:
            if canery_type == CaneryType.underflow or canery_type == CaneryType.overflow:
                self.ql.hook_mem_write(self.bo_handler, begin=canary_begin, end=canary_end)
                self.ql.hook_mem_read(self.oob_handler, begin=canary_begin, end=canary_end)
            else:
                self.ql.hook_mem_valid(self.uaf_handler, begin=canary_begin, end=canary_end)

    @staticmethod
    def bo_handler(ql, access, addr, size, value):
        """
        Called when a buffer overflow/underflow is detected.
        """
        pass

    @staticmethod
    def oob_handler(ql, access, addr, size, value):
        """
        Called when an out-of-bounds element is accessed.
        """
        pass

    @staticmethod
    def uaf_handler(ql, access, addr, size, value):
        """
        Called when a use-after-free is detected.
        """
        pass

    @staticmethod
    def bad_free_handler(ql, addr):
        """
        Called when a bad/double free is detected.
        """
        pass

    def alloc(self, size):
        chance = random.randint(1, 100)
        if chance <= self.fault_rate:
            # Fail the allocation.
            return 0

        # Add 8 bytes to the requested size so as to accomodate the canaries.
        addr = self.heap.alloc(size + 8)
        self.ql.mem.write(addr, self.canary_byte * (size + 8))

        # Install canary hooks for overflow/underflow detection.
        underflow_canary = (addr, addr + 3, CaneryType.underflow)
        self.ql.hook_mem_write(self.bo_handler, begin=underflow_canary[0], end=underflow_canary[1])
        self.ql.hook_mem_read(self.oob_handler, begin=underflow_canary[0], end=underflow_canary[1])
        self.canaries.append(underflow_canary)

        overflow_canary = (addr + 4 + size, addr + 4 + size + 3, CaneryType.overflow)
        self.ql.hook_mem_write(self.bo_handler, begin=overflow_canary[0], end=overflow_canary[1])
        self.ql.hook_mem_read(self.oob_handler, begin=overflow_canary[0], end=overflow_canary[1])
        self.canaries.append(overflow_canary)

        return (addr + 4)

    def size(self, addr):
        return self.heap.size(addr - 4)

    def free(self, addr):
        chunk = self.heap._find(addr - 4)

        if not chunk:
            self.bad_free_handler(self.ql, addr)
            return False

        # Install the UAF canary hook.
        self.ql.mem.write(addr, self.canary_byte * (chunk.size - 8))
        uaf_canary = (addr, addr + chunk.size - 8 - 1, CaneryType.uaf)
        self.ql.hook_mem_valid(self.uaf_handler, begin=uaf_canary[0], end=uaf_canary[1])
        self.canaries.append(uaf_canary)

        # Make sure the chunk won't be re-used by the underlying heap.
        self.heap.chunks.remove(chunk)
        return True

    def validate(self):
        for (canary_begin, canary_end, canery_type) in self.canaries:
            size = canary_end - canary_begin + 1
            canary = self.ql.mem.read(canary_begin, size)
            if canary.count(self.canary_byte) != len(canary):
                return False
        return True
        


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ctypes, os, struct

from unicorn.x86_const import *
from unicorn.arm_const import *

from qiling.os.const import *
from qiling.const import *
from qiling.os.macos.structs import IOExternalMethodArguments, IOExternalMethodDispatch, POINTER64

def IOConnectCallMethod(ql, selector, 
                        input_array, input_cnt, input_struct, input_struct_size,
                        output_array, output_cnt, output_struct, output_struct_size):

    if ql.os.IOKit is not True:
        ql.log.info("Must have a IOKit driver")
        return output_array, output_struct

    args_addr = ql.os.heap.alloc(ctypes.sizeof(IOExternalMethodArguments))
    ql.log.debug("Created IOExternalMethodArguments object at 0x%x" % args_addr)
    args_obj = IOExternalMethodArguments(ql, args_addr)

    if input_array is not None and input_cnt != 0:
        input_array_addr = ql.os.heap.alloc(input_cnt)
        ql.mem.write(input_array_addr, b''.join(struct.pack("<Q", x) for x in input_array))
        ql.log.debug("Created input array at 0x%x" % input_array_addr)
    else:
        input_array_addr = 0

    if input_struct is not None and input_struct_size != 0:
        input_struct_addr = ql.os.heap.alloc(input_struct_size)
        ql.mem.write(input_struct_addr, bytes(input_struct))
        ql.log.debug("Created input struct at 0x%x" % input_struct_addr)
    else:
        input_struct_addr = 0

    if output_array is not None and output_cnt != 0:
        output_array_addr = ql.os.heap.alloc(output_cnt)
        ql.mem.write(output_array_addr, b''.join(struct.pack("<Q", x) for x in output_array))
        ql.log.debug("Created output array at 0x%x" % output_array_addr)
    else:
        output_array_addr = 0

    if output_struct is not None and output_struct_size != 0:
        output_struct_addr = ql.os.heap.alloc(output_struct_size)
        ql.mem.write(output_struct_addr, bytes(output_struct))
        ql.log.debug("Created output struct at 0x%x" % output_struct_size)
    else:
        output_struct_addr = 0

    dispatch_addr = ql.os.heap.alloc(ctypes.sizeof(IOExternalMethodDispatch))
    ql.log.debug("Created IOExternalMethodDispatch object at 0x%x" % dispatch_addr)
    dispatch_obj = IOExternalMethodDispatch(ql, dispatch_addr)

    args_obj.___reservedA = 0
    args_obj._version = 2
    args_obj._selector = selector
    args_obj._asyncWakePort = 0
    args_obj._asyncReference = POINTER64(0)
    args_obj._asyncReferenceCount = 0
    args_obj._structureVariableOutputData = POINTER64(0)
    args_obj._scalarInput = POINTER64(input_array_addr)
    args_obj._scalarInputCount = input_cnt
    args_obj._structureInput = POINTER64(input_struct_addr)
    args_obj._structureInputSize = input_struct_size

    # https://stackoverflow.com/questions/45432476/sending-iokit-command-with-dynamic-length
    args_obj._structureInputDescriptor = POINTER64(0)
    args_obj._structureOutputDescriptor = POINTER64(0)
    args_obj._structureOutputDescriptorSize = 0

    args_obj._scalarOutput = POINTER64(output_array_addr)
    args_obj._scalarOutputCount = output_cnt
    args_obj._structureOutput = POINTER64(output_struct_addr)
    args_obj._structureOutputSize = output_struct_size

    args_obj.updateToMem()
    ql.log.debug("Initialized IOExternalMethodArguments object")
    ql.os.savedrip=0xffffff8000a106ba
    ql.run(begin=ql.loader.user_alloc)
    ql.os.user_object = ql.reg.rax
    ql.log.debug("Created user object at 0x%x" % ql.os.user_object)

    ql.reg.rdi = ql.os.user_object
    ql.reg.rsi = 0x1337 # owningTask
    ql.reg.rdx = 0 # securityID
    ql.reg.rcx = 0 # type
    ql.reg.r8 = 0 # properties
    ql.stack_push(0)
    ql.os.savedrip=0xffffff8000a10728
    ql.run(begin=ql.loader.user_initWithTask)
    ql.log.debug("Initialized user object")

    # TODO: Add some extra methods with correct order

    ql.reg.rdi = ql.os.user_object
    ql.reg.rsi = selector
    ql.reg.rdx = args_addr
    ql.reg.rcx = dispatch_addr
    ql.reg.r8 = ql.os.kext_object
    ql.reg.r9 = 0
    ql.os.savedrip=0xffffff8000a6e9c7
    ql.run(begin=ql.loader.user_externalMethod)

    args_obj.loadFromMem()
    output_array = args_obj.scalarOutput
    ql.log.debug("Finish IOConnectCallMethod")
    return args_obj.scalarOutput, type(output_struct).from_buffer(args_obj.structureOutput)


def gen_stub_code(ql, params, func, ret=0):
    """
    stub:
        0:  48 bf aa aa aa aa aa    movabs rdi,0xaaaaaaaaaaaaaaaa
        7:  aa aa aa
        a:  48 be bb bb bb bb bb    movabs rsi,0xbbbbbbbbbbbbbbbb
        11: bb bb bb
        14: 48 ba cc cc cc cc cc    movabs rdx,0xcccccccccccccccc
        1b: cc cc cc
        1e: 48 b9 dd dd dd dd dd    movabs rcx,0xdddddddddddddddd
        25: dd dd dd
        28: 49 b8 ee ee ee ee ee    movabs r8,0xeeeeeeeeeeeeeeee
        2f: ee ee ee
        32: 49 b9 22 22 22 22 22    movabs r9,0x2222222222222222
        39: 22 22 22
        3c: 48 b8 11 11 11 11 11    movabs rax,0x1111111111111111
        43: 11 11 11
        46: ff e0                   jmp    rax
    """
    reg_list = [b"\x48\xbf", b"\x48\xbe", b"\x48\xba", b"\x48\xb9", b"\x49\xb8", b"\x49\xb9"]
    remains = 0
    shellcode = b""
    if len(params) <= 6:
        for idx, p in enumerate(params):
            shellcode += reg_list[idx] + struct.pack("<Q", p)
    else:
        for idx, p in enumerate(params[:6]):
            shellcode += reg_list[idx] + struct.pack("<Q", p)
        remains = len(params) - 6
        for i in range(remains):
            shellcode += b"\x48\xb8" + struct.pack("<Q", params[i + 6]) + b"\x50"

    clean_stack = b""
    if remains > 0:
        """
        clean_stack:
            0:  48 81 c4 ff ff 00 00    add    rsp,0xffff
            7:  c3                      ret
        """
        clean_stack = b"\x48\x81\xc4" + struct.pack("<H", remains * 8) + b"\x00\x00\xc3"
    else:
        """
        clean_stack:
            0:  c3                      ret
        """
        clean_stack = b"\xc3"

    cleaner = ql.os.heap.alloc(len(clean_stack))
    ql.mem.write(cleaner, clean_stack)

    if func > 0:
        shellcode += b"\x48\xb8" + struct.pack("<Q", cleaner) + b"\x50"
        shellcode += b"\x48\xb8" + struct.pack("<Q", func)
        shellcode += b"\xff\xe0"
    else:
        shellcode = b"\x48\xb8" + struct.pack("<Q", ret) + b"\xc3"

    trampoline = ql.os.heap.alloc(len(shellcode))
    ql.mem.write(trampoline, shellcode)
    return trampoline

def env_dict_to_array(env_dict):
    env_list = []
    for item in env_dict:
        env_list.append(item + "=" + env_dict[item])
    return env_list


def page_align_end(addr, page_size):
    if addr % page_size == 0:
        return addr
    else:
        return int(((addr / page_size) + 1) * page_size)


def set_eflags_cf(ql, target_cf):
    ql.reg.ef = ( ql.reg.ef & 0xfffffffe ) | target_cf
    return ql.reg.ef



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from .utils import check_and_notify_protocols

def hook_EndOfExecution(ql):
    if ql.os.notify_after_module_execution(ql, len(ql.loader.modules)):
        return

    ql.loader.restore_runtime_services()

    if check_and_notify_protocols(ql):
        return

    if ql.loader.modules:
        ql.loader.execute_next_module()
    else:
        if ql.loader.unload_modules():
            return

        ql.log.info(f'No more modules to run')
        ql.emu_stop()



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os

from struct import *

from qiling.const import *
from .const import *
from .mach_port import *


def vm_shared_region_enter(ql):
    ql.mem.map(SHARED_REGION_BASE_X86_64, SHARED_REGION_SIZE_X86_64, info="[shared_region]")
    ql.macos_shared_region = True
    ql.macos_shared_region_port = MachPort(9999)        # random port name


def map_commpage(ql):
    if ql.archtype== QL_ARCH.X8664:
        addr_base = X8664_COMM_PAGE_START_ADDRESS
        addr_size = 0x100000
    elif ql.archtype== QL_ARCH.ARM64:
        addr_base = ARM64_COMM_PAGE_START_ADDRESS
        addr_size = 0x1000        
    ql.mem.map(addr_base, addr_size, info="[commpage]")
    time_lock_slide = 0x68
    ql.mem.write(addr_base+time_lock_slide, ql.pack32(0x1))


# reference to osfmk/mach/shared_memory_server.h
class SharedFileMappingNp:

    def __init__(self, ql):
        self.size = 32
        self.ql = ql
    
    def read_mapping(self, addr):
        content = self.ql.mem.read(addr, self.size)
        self.sfm_address = unpack("<Q", self.ql.mem.read(addr, 8))[0]
        self.sfm_size = unpack("<Q", self.ql.mem.read(addr + 8, 8))[0]
        self.sfm_file_offset = unpack("<Q", self.ql.mem.read(addr + 16, 8))[0]
        self.sfm_max_prot = unpack("<L", self.ql.mem.read(addr + 24, 4))[0]
        self.sfm_init_prot = unpack("<L", self.ql.mem.read(addr + 28, 4))[0]

        self.ql.log.debug("[ShareFileMapping]: addr: 0x{:X}, size: 0x{:X}, fileOffset:0x{:X}, maxProt: {}, initProt: {}".format(
            self.sfm_address, self.sfm_size, self.sfm_file_offset, self.sfm_max_prot, self.sfm_init_prot
            ))


# reference to bsd/sys/proc_info.h
class ProcRegionWithPathInfo():

    def __init__(self, ql):
        self.ql = ql
        pass
    
    def set_path(self, path):
        self.vnode_info_path_vip_path = path

    def write_info(self, addr):
        addr += 248
        self.ql.mem.write(addr, self.vnode_info_path_vip_path)


# virtual FS
# Only have some basic func now 
# tobe completed
class FileSystem():

    def __init__(self, ql):
        self.ql = ql
        self.base_path = ql.rootfs

    def get_common_attr(self, path, cmn_flags):
        real_path = self.vm_to_real_path(path)
        if not os.path.exists(real_path):
            return None
        attr = b''
        file_stat = os.stat(real_path)
        filename = ""

        if cmn_flags & ATTR_CMN_NAME != 0:
            filename = path.split("/")[-1]
            filename_len = len(filename) + 1        # add \0
            attr += pack("<L", filename_len)
            self.ql.log.debug("FileName :{}, len:{}".format(filename, filename_len))

        if cmn_flags & ATTR_CMN_DEVID != 0:
            attr += pack("<L", file_stat.st_dev)
            self.ql.log.debug("DevID: {}".format(file_stat.st_dev))

        if cmn_flags & ATTR_CMN_OBJTYPE != 0:
            if os.path.isdir(path):
                attr += pack("<L", VDIR)
                self.ql.log.debug("ObjType: DIR")
            elif os.path.islink(path):
                attr += pack("<L", VLINK)
                self.ql.log.debug("ObjType: LINK")
            else:
                attr += pack("<L", VREG)
                self.ql.log.debug("ObjType: REG")
            
        if cmn_flags & ATTR_CMN_OBJID != 0:
            attr += pack("<Q", file_stat.st_ino)
            self.ql.log.debug("VnodeID :{}".format(file_stat.st_ino))

        # at last, add name 
        if cmn_flags & ATTR_CMN_NAME != 0:
            name_offset = len(attr) + 4
            attr = pack("<L", name_offset) + attr
            attr += filename.encode("utf8")
            attr += b'\x00'
        
        self.ql.log.debug("Attr : {}".format(attr))
    
        return attr

    def vm_to_real_path(self, vm_path):
        if not vm_path:
            return None
        if vm_path[0] == '/':
            # abs path 
            return os.path.join(self.base_path, vm_path[1:])
        else:
            # rel path
            return os.path.join(self.base_path, vm_path)

    def open(self, path, open_flags, open_mode):

        real_path = self.vm_to_real_path(path)
        
        if real_path:
            return os.open(real_path, open_flags, open_mode)
        else:
            return None

    def isexists(self, path):
        real_path = self.vm_to_real_path(path)
        return os.path.exists(real_path)



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.arch.x86_const import *

def ql_syscall_clock_gettime(ql, clock_gettime_clock_id, clock_gettime_timespec, *args, **kw):
    ql.log.info("clock_gettime()")
    return 0


def ql_syscall_sysarch(ql, op, parms, *args, **kw):
    """
    wild guess, of cause not working
    """

    regreturn = 0
    ql.GS_SEGMENT_ADDR = 0x6000
    ql.GS_SEGMENT_SIZE = 0x8000


    #ql.mem.map(ql.GS_SEGMENT_ADDR, ql.GS_SEGMENT_SIZE)
    #ql.reg.msr(GSMSR, ql.GS_SEGMENT_ADDR)
    ql.reg.msr(FSMSR, parms)

    #op_buf = ql.pack32(op)
    #ql.mem.write(parms, op_buf)

    return regreturn



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

# BOOL WINAPI SetConsoleTitle(
#   _In_ LPCTSTR lpConsoleTitle
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"lpConsoleTitle": WSTRING})
def hook_SetConsoleTitleW(ql, address, params):
    return 1


# BOOL WINAPI GetConsoleScreenBufferInfo(
#   _In_  HANDLE                      hConsoleOutput,
#   _Out_ PCONSOLE_SCREEN_BUFFER_INFO lpConsoleScreenBufferInfo
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"hConsoleOutput": HANDLE, "lpConsoleScreenBufferInfo": POINTER})
def hook_GetConsoleScreenBufferInfo(ql, address, params):
    # Todo should we really?
    return 1


# BOOL WINAPI SetConsoleTextAttribute(
#   _In_ HANDLE hConsoleOutput,
#   _In_ WORD   wAttributes
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"hConsoleOutput": HANDLE, "wAttributes": INT})
def hook_SetConsoleTextAttribute(ql, address, params):
    return 1



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from .UefiBaseType import *
from .ProcessorBind import *

EFI_SMRAM_OPEN					= 0x00000001
EFI_SMRAM_CLOSED				= 0x00000002
EFI_SMRAM_LOCKED				= 0x00000004
EFI_CACHEABLE					= 0x00000008
EFI_ALLOCATED					= 0x00000010
EFI_NEEDS_TESTING				= 0x00000020
EFI_NEEDS_ECC_INITIALIZATION	= 0x00000040

class EFI_SMRAM_DESCRIPTOR(STRUCT):
	_fields_ = [
		('PhysicalStart',	EFI_PHYSICAL_ADDRESS),
		('CpuStart',		EFI_PHYSICAL_ADDRESS),
		('PhysicalSize',	UINT64),
		('RegionState',		UINT64)
	]

__all__ = [
	'EFI_SMRAM_DESCRIPTOR',
	'EFI_SMRAM_OPEN',
	'EFI_SMRAM_CLOSED',
	'EFI_SMRAM_LOCKED',
	'EFI_CACHEABLE',
	'EFI_ALLOCATED',
	'EFI_NEEDS_TESTING',
	'EFI_NEEDS_ECC_INITIALIZATION'
]



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_sendfile64(ql, sendfile64_out_fd, sendfile64_in_fd, sendfile64_offest, sendfile64_count, *args, **kw):
    if sendfile64_out_fd >= 0 and sendfile64_out_fd < 256 and sendfile64_in_fd >= 0 and sendfile64_in_fd < 256:
        ql.os.fd[sendfile64_in_fd].lseek(ql.unpack32(ql.mem.read(sendfile64_offest, 4)))
        buf = ql.os.fd[sendfile64_in_fd].read(sendfile64_count)
        regreturn = ql.os.fd[sendfile64_out_fd].write(buf)
    else:
        regreturn = -1
    return regreturn



#!/usr/bin/env python3
#
# Python setup for Qiling framework

import sys, os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
gb = {}
with open(os.path.join(here, "qiling", "__version__.py"), "r+") as f:
    exec(f.read(), gb)

VERSION = gb['__version__']

requirements = [
    "capstone>=4.0.1",
    "unicorn>=1.0.2",
    "pefile>=2019.4.18",
    "python-registry>=1.3.1",
    "keystone-engine>=0.9.2",
    "pyelftools>=0.26",
    "gevent>=20.9.0"
]

with open("README.md", "r", encoding="utf-8") as ld:
    long_description = ld.read()

if "win32" in sys.platform:
    requirements += ["windows-curses>=2.1.0"]

setup(
    name='qiling',
    version=VERSION,

    description='Qiling is an advanced binary emulation framework that cross-platform-architecture',
    url='http://qiling.io',
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer='KaiJern Lau (xwings)',
    maintainer_email='info@qiling.io',

    license='GPLv2',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    keywords='qiling binary emulator framework malware analysis UEFI IoT',

    packages=find_packages(),
    scripts=['qltool'],
    include_package_data=True,
    install_requires=requirements,
)



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_ioctl(ql, ioctl_fd, ioctl_cmd, ioctl_arg, *args, **kw):
    TCGETS = 0x5401
    TIOCGWINSZ = 0x5413
    TIOCSWINSZ = 0x5414
    TCSETSW = 0x5403

    SIOCGIFADDR = 0x8915
    SIOCGIFNETMASK = 0x891b


    def ioctl(fd, cmd, arg):
    # Stub for 'ioctl' syscall
    # Return the list of element to pack back depending on target ioctl
    #If the ioctl is disallowed, return False

        ioctl_allowed = None # list of (fd, cmd), None value for wildcard
        ioctl_disallowed = None # list of (fd, cmd), None value for wildcard

        ioctl_allowed = [
        (0, TCGETS),
        (0, TIOCGWINSZ),
        (0, TIOCSWINSZ),
        (1, TCGETS),
        (1, TIOCGWINSZ),
        (1, TIOCSWINSZ),
        ]

        ioctl_disallowed = [
        (2, TCGETS),
        (0, TCSETSW),
        ]

        allowed = False
        disallowed = False

        for test in [(fd, cmd), (None, cmd), (fd, None)]:
            if test in ioctl_allowed:
                allowed = True
            if test in ioctl_disallowed:
                disallowed = True
        if allowed and disallowed:
            raise ValueError("fd: %x, cmd: %x is allowed and disallowed" % (fd, cmd))

        if allowed:

            if cmd == TCGETS:
                return 0, 0, 0, 0
            elif cmd == TIOCGWINSZ:
            # struct winsize
            # {
            #   unsigned short ws_row;	/* rows, in characters */
            #   unsigned short ws_col;	/* columns, in characters */
            #   unsigned short ws_xpixel;	/* horizontal size, pixels */
            #   unsigned short ws_ypixel;	/* vertical size, pixels */
            # };
                return 1000, 360, 1000, 1000
            elif cmd == TIOCSWINSZ:
                # Ignore it
                return
            else:
                raise RuntimeError("Not implemented")
        elif disallowed:
            return False
        else:
            raise KeyError("Unknown ioctl fd:%x cmd:%x" % (fd, cmd))

    if isinstance(ql.os.fd[ioctl_fd], ql_socket) and (ioctl_cmd == SIOCGIFADDR or ioctl_cmd == SIOCGIFNETMASK):
        try:
            tmp_arg = ql.mem.read(ioctl_arg, 64)
            ql.log.debug("query network card : %s" % tmp_arg)
            data = ql.os.fd[ioctl_fd].ioctl(ioctl_cmd, bytes(tmp_arg))
            ql.mem.write(ioctl_arg, data)
            regreturn = 0
        except:
            regreturn = -1
    else:
        try:
            info = ioctl(ioctl_fd, ioctl_cmd, ioctl_arg)
            if ioctl_cmd == TCGETS:
                data = struct.pack("BBBB", *info)
                ql.mem.write(ioctl_arg, data)
            elif ioctl_cmd == TIOCGWINSZ:
                data = struct.pack("HHHH", *info)
                ql.mem.write(ioctl_arg, data)
            else:
                return
            regreturn = 0
        except :
            regreturn = -1

    ql.log.debug("ioctl(0x%x, 0x%x, 0x%x) = %d" % (ioctl_fd, ioctl_cmd, ioctl_arg, regreturn))
    return regreturn



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time


from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

# BOOL IsDebuggerPresent();
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_IsDebuggerPresent(ql, address, params):
    return 0


# BOOL CheckRemoteDebuggerPresent(
#   HANDLE hProcess,
#   PBOOL  pbDebuggerPresent
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CheckRemoteDebuggerPresent(ql, address, params):
    pointer = params["pbDebuggerPresent"]
    ql.mem.write(pointer, 0x0.to_bytes(1, byteorder="little"))
    return 1


# void OutputDebugStringW(
#   LPCWSTR lpOutputString
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_OutputDebugStringW(ql, address, params):
    string = params["lpOutputString"]
    ql.log.info('OutputDebugStringW: "%s"' % (string.encode()))
    return 0


# void OutputDebugStringA(
#  LPCSTR lpOutputString
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_OutputDebugStringA(ql, address, params):
    ql.log.info('OutputDebugStringA: "%s"' % (params['lpOutputString']))
    return 0



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.arch.x86 import *
from qiling.os.posix.posix import QlOsPosix
from .const import *
from .utils import *
from .futex import *
from .thread import *

class QlOsLinux(QlOsPosix):
    def __init__(self, ql):
        super(QlOsLinux, self).__init__(ql)
        self.ql = ql
        self.thread_class = None
        self.futexm = None
        self.function_hook_tmp = []
        self.fh = None
        self.user_defined_api = {}
        self.function_after_load_list = []
        self.pid = self.profile.getint("KERNEL","pid")
        self.load()

        if self.ql.archtype == QL_ARCH.X8664:
            ql_x8664_set_gs(self.ql)

    def load(self):
        self.futexm = QlLinuxFutexManagement()

        # ARM
        if self.ql.archtype== QL_ARCH.ARM:
            self.ql.arch.enable_vfp()
            self.ql.hook_intno(self.hook_syscall, 2)
            self.thread_class = QlLinuxARMThread
            ql_arm_init_get_tls(self.ql)


        # MIPS32
        elif self.ql.archtype== QL_ARCH.MIPS:      
            self.ql.hook_intno(self.hook_syscall, 17)
            self.thread_class = QlLinuxMIPS32Thread

        # ARM64
        elif self.ql.archtype== QL_ARCH.ARM64:
            self.ql.arch.enable_vfp()
            self.ql.hook_intno(self.hook_syscall, 2)
            self.thread_class = QlLinuxARM64Thread

        # X86
        elif  self.ql.archtype== QL_ARCH.X86:
            self.gdtm = GDTManager(self.ql)
            ql_x86_register_cs(self)
            ql_x86_register_ds_ss_es(self)
            self.ql.hook_intno(self.hook_syscall, 0x80)
            self.thread_class = QlLinuxX86Thread

        # X8664
        elif  self.ql.archtype== QL_ARCH.X8664:
            self.gdtm = GDTManager(self.ql)
            ql_x86_register_cs(self)
            ql_x86_register_ds_ss_es(self)
            self.ql.hook_insn(self.hook_syscall, UC_X86_INS_SYSCALL)
            # Keep test for _cc
            #self.ql.hook_insn(hook_posix_api, UC_X86_INS_SYSCALL)
            self.thread_class = QlLinuxX8664Thread
       
    def hook_syscall(self, int= None, intno= None):
        return self.load_syscall()


    def add_function_hook(self, fn, cb, userdata = None):
        self.function_hook_tmp.append((fn, cb, userdata))


    def register_function_after_load(self, function):
        if function not in self.function_after_load_list:
            self.function_after_load_list.append(function)


    def run_function_after_load(self):
        for f in self.function_after_load_list:
            f()


    def run(self):
        for function, callback, userdata in self.ql.os.function_hook_tmp:
            self.ql.os.function_hook.add_function_hook(function, callback, userdata)

        if self.ql.exit_point is not None:
            self.exit_point = self.ql.exit_point

        try:
            if self.ql.code:
                self.ql.emu_start(self.entry_point, (self.entry_point + len(self.ql.code)), self.ql.timeout, self.ql.count)
            else:
                if self.ql.multithread == True:
                    # start multithreading
                    thread_management = QlLinuxThreadManagement(self.ql)
                    self.ql.os.thread_management = thread_management
                    thread_management.run()

                else:
                    
                    if  self.ql.entry_point is not None:
                        self.ql.loader.elf_entry = self.ql.entry_point

                    elif self.ql.loader.elf_entry != self.ql.loader.entry_point:
                        self.ql.emu_start(self.ql.loader.entry_point, self.ql.loader.elf_entry, self.ql.timeout)
                        self.ql.enable_lib_patch()
                        self.run_function_after_load()
                        self.ql.loader.skip_exit_check = False
                        self.ql.write_exit_trap()

                    self.ql.emu_start(self.ql.loader.elf_entry, self.exit_point, self.ql.timeout, self.ql.count)

        except UcError:
            # TODO: this is bad We need a better approach for this
            #if self.ql.output != QL_OUTPUT.DEBUG:
            #    return
            
            self.emu_error()
            raise




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from struct import unpack

from .utils import *
from .const import *

class Header:

    def __init__(self, data):
        self.binary = data
            

class BinaryHeader(Header):

    def __init__(self, data):
        super().__init__(data)
        FR = FileReader(data)
        self.magic          = unpack("<L", FR.read(4))[0]
        self.cpu_type       = unpack("<L", FR.read(4))[0]
        self.cpu_subtype    = unpack("<L", FR.read(4))[0]
        self.file_type      = unpack("<L", FR.read(4))[0]
        self.lc_num         = unpack("<L", FR.read(4))[0]
        self.lc_size        = unpack("<L", FR.read(4))[0]
        self.flags          = unpack("<L", FR.read(4))[0]
        
        if self.magic in MAGIC_64:
            self.reserved = unpack("<L", FR.read(4))[0]
        self.header_size = FR.offset

    #def __str__(self):
        # return ("magic : 0x%X, cputype: 0x%X, subType: 0x%X, FileType: 0x%X, lc num: 0x%X, lc size: 0x%X, flags: 0x%X" % (self.magic, 
        #     self.cpu_type, self.cpu_subtype, self.file_type, self.lc_num, self.lc_size, self.flags))


class FatHeader(Header):

    def __init__(self, data):
        super().__init__(data)
        FR = FileReader(data)
        FR.setOffset(4)
        self.binarys = []
        self.arch_num = unpack(">L", FR.read(4))[0]
        for i in range(self.arch_num):
            FI = FatInfo(FR.read(4 * 5))
            self.binarys.append(FI)

    def getBinary(self, arch):

        for item in self.binarys:
            if item.cpu_type == CPU_TYPE_X8664:
                return item
            elif item.cpu_type == CPU_TYPE_ARM64:
                return item 
        return None

class FatInfo:
    def __init__(self, data):
        FR = FileReader(data)
        self.cpu_type       = unpack(">L", FR.read(4))[0]
        self.cpu_subtype    = unpack(">L", FR.read(4))[0]
        self.offset         = unpack(">L", FR.read(4))[0]
        self.size           = unpack(">L", FR.read(4))[0]
        self.align          = 2 ** unpack(">L", FR.read(4))[0]
    
    # def __str__(self):
    #     return ("CPU 0x%X, CPU subtype 0x%X, offset 0x%X, size 0x%X, align %d" %(
    #         self.cpu_type, self.cpu_subtype, self.offset, self.size, self.align
    #     ))


#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import traceback

from unicorn import *
from unicorn.x86_const import *
from unicorn.arm64_const import *


from qiling.arch.x86 import *
from qiling.const import *
from qiling.os.const import *
from qiling.os.posix.posix import QlOsPosix
from .const import *
from qiling.os.macos import macos
from qiling.os.macos.structs import *
from qiling.os.macos.events.macos_structs import *
from qiling.os.macos.events.macos import *
from qiling.os.macos.events.macos_policy import *

class QlOsMacos(QlOsPosix):
    def __init__(self, ql):
        super(QlOsMacos, self).__init__(ql)
        self.ql = ql
        self.ql.counter = 0
        self.user_defined_api = {}
        self.ev_manager = QlMacOSEvManager(self.ql)
        self.policy_manager = QlMacOSPolicy(self.ql, self.ev_manager)
        self.RUN = True
        self.hook_ret = {}
        self.pid = self.profile.getint("KERNEL","pid")
        self.load()


    # load MacOS driver
    def load_kext(self):
        from qiling.os.macos.events.macos_structs import mac_policy_list_t
        from qiling.os.macos.structs import list_head
        self.heap.clear()

        # Setup mac_policy_list
        kern_mac_policy_list = self.ql.loader.kernel_extrn_symbols_detail[b"_mac_policy_list"]["n_value"]
        mac_policy_list_addr = self.heap.alloc(ctypes.sizeof(mac_policy_list_t))
        self.mac_policy_list = mac_policy_list_t(self, mac_policy_list_addr)
        self.ql.mem.write(kern_mac_policy_list, struct.pack("<Q", mac_policy_list_addr))

        # Add initial process to allproc in kernel
        allproc = self.ql.loader.kernel_extrn_symbols_detail[b"_allproc"]["n_value"]
        self.ev_manager.set_allproc(allproc)
        self.ev_manager.add_process(0, "head")
        self.ev_manager.add_process(0x1337, "demigod")
        self.ev_manager.add_process(1, "tail")

        if self.ql.loader.IOKit is True: # Handle IOKit driver
            self.ql.stack_push(0)
            self.savedrip=0xffffff8000a163bd
            self.ql.run(begin=self.ql.loader.kext_alloc)
            self.kext_object = self.ql.reg.rax
            self.ql.log.debug("Created kext object at 0x%x" % self.kext_object)

            self.ql.reg.rdi = self.kext_object
            self.ql.reg.rsi = 0 # NULL option
            self.savedrip=0xffffff8000a16020
            self.ql.run(begin=self.ql.loader.kext_init)
            if self.ql.reg.rax == 0:
                self.ql.log.debug("Failed to initialize kext object")
                return
            self.ql.log.debug("Initialized kext object")

            self.ql.reg.rdi = self.kext_object
            # FIXME Determine provider for kext
            self.ql.reg.rsi = 0 # ?
            self.savedrip=0xffffff8000a16102
            self.ql.run(begin=self.ql.loader.kext_attach)
            if self.ql.reg.rax == 0:
                self.ql.log.debug("Failed to attach kext object")
                return
            self.ql.log.debug("Attached kext object 1st time")

            self.ql.reg.rdi = self.kext_object
            self.ql.reg.rdi = 0
            # FIXME Determine provider for kext
            self.ql.reg.rsi = 0 # ?
            tmp = self.heap.alloc(8)
            self.ql.reg.rdx = tmp
            self.savedrip=0xffffff8000a16184
            self.ql.run(begin=self.ql.loader.kext_probe)
            self.heap.free(tmp)
            self.ql.log.debug("Probed kext object")

            self.ql.reg.rdi = self.kext_object
            # FIXME Determine provider for kext
            self.ql.reg.rsi = 0 # ?
            self.savedrip=0xffffff8000a16198
            self.ql.run(begin=self.ql.loader.kext_detach)
            self.ql.log.debug("Detached kext object")

            self.ql.reg.rdi = self.kext_object
            # FIXME Determine provider for kext
            self.ql.reg.rsi = 0 # ?
            self.savedrip=0xffffff8000a168a3
            self.ql.run(begin=self.ql.loader.kext_attach)
            if self.ql.reg.rax == 0:
                self.ql.log.debug("Failed to attach kext object")
                return
            self.ql.log.debug("Attached kext object 2nd time")

            self.ql.reg.rdi = self.kext_object
            # FIXME Determine provider for kext
            self.ql.reg.rsi = 0 # ?
            self.savedrip=0xffffff8000a168ed
            self.ql.run(begin=self.ql.loader.kext_start)
        else:
            from qiling.os.macos.structs import kmod_info_t, POINTER64
            kmod_info_addr = self.heap.alloc(ctypes.sizeof(kmod_info_t))
            self.ql.log.debug("Created fake kmod_info at 0x%x" % kmod_info_addr)
            kmod_info = kmod_info_t(self.ql, kmod_info_addr)

            # OSKext.cpp:562
            kmod_info.next = POINTER64(0)
            kmod_info.info_version = 1
            kmod_info.id = 1
            kmod_info.name = self.ql.loader.plist["CFBundleIdentifier"].encode()
            kmod_info.version = self.ql.loader.plist["CFBundleVersion"].encode()
            kmod_info.reference_count = 0
            kmod_info.reference_list = POINTER64(0)
            kmod_info.address = POINTER64(self.ql.loader.slide)
            kmod_info.size = self.ql.loader.kext_size
            kmod_info.hdr_size = self.ql.loader.macho_file.header.header_size
            kmod_info.start = POINTER64(self.ql.loader.kext_start)
            kmod_info.stop = POINTER64(self.ql.loader.kext_stop)

            kmod_info.updateToMem()
            self.ql.log.debug("Initialized kmod_info")

            self.ql.reg.rdi = kmod_info_addr
            self.ql.reg.rsi = 0
            self.savedrip=0xffffff80009c2c16
            self.ql.run(begin=self.ql.loader.kext_start)


    def load(self):
        if self.ql.code:
            return

        if self.ql.archtype== QL_ARCH.ARM64:
            self.ql.arch.enable_vfp()
            self.ql.hook_intno(self.hook_syscall, 2)
            self.ql.hook_intno(self.hook_sigtrap, 7)

        elif self.ql.archtype== QL_ARCH.X8664:
            self.ql.hook_insn(self.hook_syscall, UC_X86_INS_SYSCALL)
            self.gdtm = GDTManager(self.ql)
            ql_x86_register_cs(self)
            ql_x86_register_ds_ss_es(self)

    
    def hook_syscall(self, intno= None, int = None):
        return self.load_syscall()


    def hook_sigtrap(self, intno= None, int = None):
        self.ql.log.info("Trap Found")
        self.emu_error()
        exit(1)


    def run(self):
        #save initial stack pointer, so we can see if stack is balanced when
        #this function return at the end
        if self.ql.loader.kext_name and self.savedrip is not None:
            """
            Use following code to extract saved rip from method:
            {
                ...
                unsigned long rbp_register;
                __asm__ volatile ("mov %%rbp, %0" : "=r" (rbp_register));
                unsigned char *saved_rip = (unsigned char *) (*(unsigned long *)(rbp_register + sizeof(void *)));
                ...
            }
            """
            self.ql.stack_push(self.savedrip)
            def callback_ret(ql):
                ql.reg.arch_pc = 0
                
            if self.savedrip not in self.hook_ret:
                tmp = self.ql.hook_address(callback_ret, self.savedrip)
                self.hook_ret[self.savedrip] = tmp
        elif self.ql.loader.kext_name:
            self.ql.stack_push(0)

        if self.ql.exit_point is not None:
            self.exit_point = self.ql.exit_point

        if  self.ql.entry_point is not None:
                self.ql.loader.entry_point = self.ql.entry_point    

        try:
            if self.ql.code:
                self.ql.emu_start(self.entry_point, (self.entry_point + len(self.ql.code)), self.ql.timeout, self.ql.count)
            
            else:
                self.ql.emu_start(self.ql.loader.entry_point, self.exit_point, self.ql.timeout, self.ql.count)
        except UcError:
            self.RUN = False

            # TODO: This is for kext, we need to fix it later
            if self.ql.output != QL_OUTPUT.DEBUG:
                return
            
            self.emu_error()
            raise

        self.RUN = False

        if self.ql._internal_exception != None:
            raise self.ql._internal_exception


#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os, pefile, pickle, secrets, string, sys, traceback

from unicorn.x86_const import *


from qiling.os.windows.utils import *
from qiling.os.windows.structs import *
from qiling.exception import *
from qiling.const import *
from qiling.arch.x86_const import *
from .loader import QlLoader
from qiling.os.memory import QlMemoryHeap


class QlPeCacheEntry:
    def __init__(self, data, cmdlines, import_symbols, import_table):
        self.data = data
        self.cmdlines = cmdlines
        self.import_symbols = import_symbols
        self.import_table = import_table

# A default simple cache implementation
class QlPeCache:
    def create_filename(self, path, address):
        return path + ".%x.cache2" % address

    def restore(self, path, address):
        fcache = self.create_filename(path, address)
        # pickle file cannot be outdated
        if os.path.exists(fcache) and os.stat(fcache).st_mtime > os.stat(path).st_mtime:
            with open(fcache, "rb") as fcache_file:
                return QlPeCacheEntry(*pickle.load(fcache_file))
        return None

    def save(self, path, address, entry):
        fcache = self.create_filename(path, address)
        data = (entry.data, entry.cmdlines, entry.import_symbols, entry.import_table)
        # cache this dll file
        with open(fcache, "wb") as fcache_file:
            pickle.dump(data, fcache_file)


class Process():
    def __init__(self, ql):
        self.ql = ql

    def align(self, size, unit):
        return (size // unit + (1 if size % unit else 0)) * unit

    def load_dll(self, dll_name, driver=False):
        dll_name = dll_name.decode()

        self.ql.dlls = os.path.join("Windows", "System32")

        if 'C:\\' in dll_name.upper():
            path = self.ql.os.transform_to_real_path(dll_name)
            dll_name = path_leaf(dll_name)
        else:
            dll_name = dll_name.lower()
            if not is_file_library(dll_name):
                dll_name = dll_name + ".dll"
            path = os.path.join(self.ql.rootfs, self.ql.dlls, dll_name)

        if not os.path.exists(path):
            raise QlErrorFileNotFound("Cannot find dll in %s" % path)

        # If the dll is already loaded
        if dll_name in self.dlls:
            return self.dlls[dll_name]
        else:
            self.dlls[dll_name] = self.dll_last_address

        self.ql.log.info("Loading %s to 0x%x" % (path, self.dll_last_address))

        if self.libcache:
            cached = self.libcache.restore(path, self.dll_last_address)
        else:
            cached = None

        if cached:
            data = cached.data
            import_symbols = cached.import_symbols
            import_table = cached.import_table
            for entry in cached.cmdlines:
                self.set_cmdline(entry['name'], entry['address'], data)
            self.ql.log.info("Loaded %s from cache" % path)
        else:
            dll = pefile.PE(path, fast_load=True)
            dll.parse_data_directories()
            warnings = dll.get_warnings()
            if warnings:
                self.ql.log.warning(f'Warnings while loading {path}:')
                for warning in warnings:
                    self.ql.log.warning(f' - {warning}')
            data = bytearray(dll.get_memory_mapped_image())
            cmdlines = []

            import_symbols = {}
            import_table = {}
            for entry in dll.DIRECTORY_ENTRY_EXPORT.symbols:
                import_symbols[self.dll_last_address + entry.address] = {"name": entry.name,
                                                                              "ordinal": entry.ordinal,
                                                                              "dll": dll_name.split('.')[0]
                                                                              }
                if entry.name:
                    import_table[entry.name] = self.dll_last_address + entry.address
                import_table[entry.ordinal] = self.dll_last_address + entry.address
                cmdline_entry = self.set_cmdline(entry.name, entry.address, data)
                if cmdline_entry:
                    cmdlines.append(cmdline_entry)

            if self.libcache:
                cached = QlPeCacheEntry(data, cmdlines, import_symbols, import_table)
                self.libcache.save(path, self.dll_last_address, cached)
                self.ql.log.info("Cached %s" % path)

        # Add dll to IAT
        try:
            self.import_address_table[dll_name] = import_table
        except Exception as ex:
            self.ql.log.exception(f'Unable to add {dll_name} to IAT')

        try:
            self.import_symbols.update(import_symbols)
        except Exception as ex:
            self.ql.log.exception(f'Unable to add {dll_name} import symbols')

        dll_base = self.dll_last_address
        dll_len = self.ql.mem.align(len(bytes(data)), 0x1000)
        self.dll_size += dll_len
        self.ql.mem.map(dll_base, dll_len, info=dll_name)
        self.ql.mem.write(dll_base, bytes(data))
        self.dll_last_address += dll_len

        # if this is NOT a driver, add dll to ldr data
        if not driver:
            self.add_ldr_data_table_entry(dll_name)

        # add DLL to coverage images
        self.images.append(self.coverage_image(dll_base, dll_base+dll_len, path))

        self.ql.log.info("Done with loading %s" % path)

        return dll_base


    def _alloc_cmdline(self, wide):
        addr = self.ql.os.heap.alloc(len(self.cmdline) * (2 if wide else 1))
        packed_addr = self.ql.pack(addr)
        return addr, packed_addr

    def set_cmdline(self, name, address, memory):
        cmdline_entry = None
        if name == b"_acmdln":
            addr, packed_addr = self._alloc_cmdline(wide=False)
            cmdline_entry = {"name": name, "address": address}
            memory[address:address + self.ql.pointersize] = packed_addr
            self.ql.mem.write(addr, self.cmdline)
        elif name == b"_wcmdln":
            addr, packed_addr = self._alloc_cmdline(wide=True)
            cmdline_entry = {"name": name, "address": address}
            memory[address:address + self.ql.pointersize] = packed_addr
            encoded = self.cmdline.decode('ascii').encode('UTF-16LE')
            self.ql.mem.write(addr, encoded)

        return cmdline_entry

    def init_tib(self):
        if self.ql.archtype == QL_ARCH.X86:
            teb_addr = self.structure_last_addr
        else:
            gs = self.structure_last_addr
            self.structure_last_addr += 0x30
            teb_addr = self.structure_last_addr

        self.ql.log.info("TEB addr is 0x%x" %teb_addr)

        teb_size = len(TEB(self.ql).bytes())
        teb_data = TEB(
            self.ql,
            base=teb_addr,
            peb_address=teb_addr + teb_size,
            stack_base=self.stack_address + self.stack_size,
            stack_limit=self.stack_size,
            Self=teb_addr)

        self.ql.mem.write(teb_addr, teb_data.bytes())

        self.structure_last_addr += teb_size
        if self.ql.archtype == QL_ARCH.X8664:
            # TEB
            self.ql.mem.write(gs + 0x30, self.ql.pack64(teb_addr))
            # PEB
            self.ql.mem.write(gs + 0x60, self.ql.pack64(teb_addr + teb_size))

        self.TEB = self.ql.TEB = teb_data

    def init_peb(self):
        peb_addr = self.structure_last_addr

        self.ql.log.info("PEB addr is 0x%x" % peb_addr)

        # we must set an heap, will try to retrieve this value. Is ok to be all \x00
        process_heap = self.ql.os.heap.alloc(0x100)
        peb_data = PEB(self.ql, base=peb_addr, process_heap=process_heap,
                       number_processors=self.ql.os.profile.getint("HARDWARE",
                                                                   "number_processors"))
        peb_data.LdrAddress = peb_addr + peb_data.size
        peb_data.write(peb_addr)
        self.structure_last_addr += peb_data.size
        self.PEB = self.ql.PEB = peb_data

    def init_ldr_data(self):
        ldr_addr = self.structure_last_addr
        ldr_size = len(LdrData(self.ql).bytes())
        ldr_data = LdrData(
            self.ql,
            base=ldr_addr,
            in_load_order_module_list={
                'Flink': ldr_addr + 2 * self.ql.pointersize,
                'Blink': ldr_addr + 2 * self.ql.pointersize
            },
            in_memory_order_module_list={
                'Flink': ldr_addr + 4 * self.ql.pointersize,
                'Blink': ldr_addr + 4 * self.ql.pointersize
            },
            in_initialization_order_module_list={
                'Flink': ldr_addr + 6 * self.ql.pointersize,
                'Blink': ldr_addr + 6 * self.ql.pointersize
            }
        )
        self.ql.mem.write(ldr_addr, ldr_data.bytes())
        self.structure_last_addr += ldr_size
        self.LDR = self.ql.LDR = ldr_data

    def add_ldr_data_table_entry(self, dll_name):
        dll_base = self.dlls[dll_name]
        path = "C:\\Windows\\System32\\" + dll_name
        ldr_table_entry_size = len(LdrDataTableEntry(self.ql).bytes())
        base = self.ql.os.heap.alloc(ldr_table_entry_size)
        ldr_table_entry = LdrDataTableEntry(self.ql,
                                            base=base,
                                            in_load_order_links={'Flink': 0, 'Blink': 0},
                                            in_memory_order_links={'Flink': 0, 'Blink': 0},
                                            in_initialization_order_links={'Flink': 0, 'Blink': 0},
                                            dll_base=dll_base,
                                            entry_point=0,
                                            full_dll_name=path,
                                            base_dll_name=dll_name)

        # Flink
        if len(self.ldr_list) == 0:
            flink = self.LDR
            ldr_table_entry.InLoadOrderLinks['Flink'] = flink.InLoadOrderModuleList['Flink']
            ldr_table_entry.InMemoryOrderLinks['Flink'] = flink.InMemoryOrderModuleList['Flink']
            ldr_table_entry.InInitializationOrderLinks['Flink'] = flink.InInitializationOrderModuleList['Flink']

            flink.InLoadOrderModuleList['Flink'] = ldr_table_entry.base
            flink.InMemoryOrderModuleList['Flink'] = ldr_table_entry.base + 2 * self.ql.pointersize
            flink.InInitializationOrderModuleList['Flink'] = ldr_table_entry.base + 4 * self.ql.pointersize

        else:
            flink = self.ldr_list[-1]
            ldr_table_entry.InLoadOrderLinks['Flink'] = flink.InLoadOrderLinks['Flink']
            ldr_table_entry.InMemoryOrderLinks['Flink'] = flink.InMemoryOrderLinks['Flink']
            ldr_table_entry.InInitializationOrderLinks['Flink'] = flink.InInitializationOrderLinks['Flink']

            flink.InLoadOrderLinks['Flink'] = ldr_table_entry.base
            flink.InMemoryOrderLinks['Flink'] = ldr_table_entry.base + 2 * self.ql.pointersize
            flink.InInitializationOrderLinks['Flink'] = ldr_table_entry.base + 4 * self.ql.pointersize

        # Blink
        blink = self.LDR
        ldr_table_entry.InLoadOrderLinks['Blink'] = blink.InLoadOrderModuleList['Blink']
        ldr_table_entry.InMemoryOrderLinks['Blink'] = blink.InMemoryOrderModuleList['Blink']
        ldr_table_entry.InInitializationOrderLinks['Blink'] = blink.InInitializationOrderModuleList['Blink']

        blink.InLoadOrderModuleList['Blink'] = ldr_table_entry.base
        blink.InMemoryOrderModuleList['Blink'] = ldr_table_entry.base + 2 * self.ql.pointersize
        blink.InInitializationOrderModuleList['Blink'] = ldr_table_entry.base + 4 * self.ql.pointersize

        self.ql.mem.write(flink.base, flink.bytes())
        self.ql.mem.write(blink.base, blink.bytes())
        self.ql.mem.write(ldr_table_entry.base, ldr_table_entry.bytes())

        self.ldr_list.append(ldr_table_entry)

    def init_exports(self):
        if self.ql.code:
            return
        if self.pe.OPTIONAL_HEADER.DATA_DIRECTORY[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_EXPORT']].VirtualAddress != 0:
            # Do a full load if IMAGE_DIRECTORY_ENTRY_EXPORT is present so we can load the exports
            self.pe.full_load()
        else:
            return

        try:
            # parse directory entry export
            dll_name = os.path.basename(self.path)
            self.import_address_table[dll_name] = {} 
            for entry in self.pe.DIRECTORY_ENTRY_EXPORT.symbols:
                self.export_symbols[self.pe_image_address + entry.address] = {'name': entry.name, 'ordinal': entry.ordinal}
                self.import_address_table[dll_name][entry.name] = self.pe_image_address + entry.address
                self.import_address_table[dll_name][entry.ordinal] = self.pe_image_address + entry.address
        except:
            self.ql.log.info('Failed to load exports for %s:\n%s' % (self.ql.argv, traceback.format_exc()))

    def init_driver_object(self):
        # PDRIVER_OBJECT DriverObject
        driver_object_addr = self.structure_last_addr
        self.ql.log.info("Driver object addr is 0x%x" %driver_object_addr)

        if self.ql.archtype == QL_ARCH.X86:
            self.driver_object = DRIVER_OBJECT32(self.ql, driver_object_addr)
        elif self.ql.archtype == QL_ARCH.X8664:
            self.driver_object = DRIVER_OBJECT64(self.ql, driver_object_addr)

        driver_object_size = ctypes.sizeof(self.driver_object)
        self.ql.mem.write(driver_object_addr, bytes(self.driver_object))
        self.structure_last_addr += driver_object_size
        self.ql.driver_object_address = driver_object_addr


    def init_registry_path(self):
        # PUNICODE_STRING RegistryPath
        regitry_path_addr = self.structure_last_addr
        self.ql.log.info("Registry path addr is 0x%x" %regitry_path_addr)

        if self.ql.archtype == QL_ARCH.X86:
            regitry_path_data = UNICODE_STRING32(0, 0, regitry_path_addr)
        elif self.ql.archtype == QL_ARCH.X8664:
            regitry_path_data = UNICODE_STRING64(0, 0, regitry_path_addr)

        regitry_path_size = ctypes.sizeof(regitry_path_data)
        self.ql.mem.write(regitry_path_addr, bytes(regitry_path_data))
        self.structure_last_addr += regitry_path_size
        self.ql.regitry_path_address = regitry_path_addr


    def init_eprocess(self):
        addr = self.structure_last_addr
        self.ql.log.info("EPROCESS is is 0x%x" %addr)


        if self.ql.archtype == QL_ARCH.X86:
            self.eprocess_object = EPROCESS32(self.ql, addr)
        elif self.ql.archtype == QL_ARCH.X8664:
            self.eprocess_object = EPROCESS64(self.ql, addr)            

        size = ctypes.sizeof(self.eprocess_object)
        self.ql.mem.write(addr, bytes(self.driver_object))
        self.structure_last_addr += size
        self.ql.eprocess_address = addr


    def init_ki_user_shared_data(self):
        '''
        https://www.geoffchappell.com/studies/windows/km/ntoskrnl/structs/kuser_shared_data/index.htm

		struct information:
		https://doxygen.reactos.org/d8/dae/modules_2rostests_2winetests_2ntdll_2time_8c_source.html
        '''
        if self.ql.archtype == QL_ARCH.X86:
            KI_USER_SHARED_DATA = 0xFFDF0000
        elif self.ql.archtype == QL_ARCH.X8664:
            KI_USER_SHARED_DATA = 0xFFFFF78000000000

        self.ql.log.info("KI_USER_SHARED_DATA is 0x%x" %KI_USER_SHARED_DATA)

        shared_user_data = KUSER_SHARED_DATA()

        shared_user_data_len = self.align(ctypes.sizeof(KUSER_SHARED_DATA), 0x1000)
        self.ql.uc.mem_map(KI_USER_SHARED_DATA, shared_user_data_len)
        self.ql.mem.write(KI_USER_SHARED_DATA, bytes(shared_user_data))


class QlLoaderPE(QlLoader, Process):
    def __init__(self, ql):
        super(QlLoaderPE, self).__init__(ql)
        self.ql         = ql
        if type(self.ql.libcache) == bool:
            self.libcache = QlPeCache() if self.ql.libcache else None
        else:
            self.libcache = self.ql.libcache
        self.path       = self.ql.path
        self.is_driver  = False

    def run(self):
        self.init_dlls = [b"ntdll.dll", b"kernel32.dll", b"user32.dll"]
        self.sys_dlls = [b"ntdll.dll", b"kernel32.dll"]
        self.pe_entry_point = 0
        self.sizeOfStackReserve = 0        

        if not self.ql.code:
            self.pe = pefile.PE(self.path, fast_load=True)
            self.is_driver = self.pe.is_driver()
            if self.is_driver == True:
                self.init_dlls.append(b"ntoskrnl.exe")
                self.sys_dlls.append(b"ntoskrnl.exe")
            
        if self.ql.archtype == QL_ARCH.X86:
            self.stack_address = int(self.ql.os.profile.get("OS32", "stack_address"), 16)
            self.stack_size = int(self.ql.os.profile.get("OS32", "stack_size"), 16)
            self.image_address = int(self.ql.os.profile.get("OS32", "image_address"), 16)
            self.dll_address = int(self.ql.os.profile.get("OS32", "dll_address"), 16)
            self.entry_point = int(self.ql.os.profile.get("OS32", "entry_point"), 16)
            self.ql.os.heap_base_address = int(self.ql.os.profile.get("OS32", "heap_address"), 16)
            self.ql.os.heap_base_size = int(self.ql.os.profile.get("OS32", "heap_size"), 16)
            self.structure_last_addr = FS_SEGMENT_ADDR
        elif self.ql.archtype == QL_ARCH.X8664:
            self.stack_address = int(self.ql.os.profile.get("OS64", "stack_address"), 16)
            self.stack_size = int(self.ql.os.profile.get("OS64", "stack_size"), 16)
            self.image_address = int(self.ql.os.profile.get("OS64", "image_address"), 16)
            self.dll_address = int(self.ql.os.profile.get("OS64", "dll_address"), 16)
            self.entry_point = int(self.ql.os.profile.get("OS64", "entry_point"), 16)
            self.ql.os.heap_base_address = int(self.ql.os.profile.get("OS64", "heap_address"), 16)
            self.ql.os.heap_base_size = int(self.ql.os.profile.get("OS64", "heap_size"), 16)
            self.structure_last_addr = GS_SEGMENT_ADDR

        self.dlls = {}
        self.import_symbols = {}
        self.export_symbols = {}
        self.import_address_table = {}
        self.ldr_list = []
        self.pe_image_address = 0
        self.pe_image_address_size = 0
        self.dll_size = 0
        self.dll_last_address = self.dll_address
        # compatible with ql.__enable_bin_patch()
        self.load_address = 0
        self.ql.os.heap = QlMemoryHeap(self.ql, self.ql.os.heap_base_address, self.ql.os.heap_base_address + self.ql.os.heap_base_size)
        self.ql.os.setupComponents()
        self.ql.os.entry_point = self.entry_point
        cmdline = (str(self.ql.os.userprofile)) + "Desktop\\" + self.ql.targetname
        self.filepath = bytes(cmdline + "\x00", "utf-8")
        for arg in self.argv[1:]:
            if ' ' in arg:
                cmdline += f' "{arg}"'
            else:
                cmdline += f' {arg}'
        cmdline += "\x00"
        self.cmdline = bytes(cmdline, "utf-8")

        self.load()

    def init_thread_information_block(self):
        super().init_tib()
        super().init_peb()
        super().init_ldr_data()
        super().init_exports()

    def load(self):
        # set stack pointer
        self.ql.log.info("Initiate stack address at 0x%x " % self.stack_address)
        self.ql.mem.map(self.stack_address, self.stack_size, info="[stack]")

        if self.path and not self.ql.code:
            # for simplicity, no image base relocation
            self.pe_image_address = self.pe.OPTIONAL_HEADER.ImageBase
            self.pe_image_address_size = self.ql.mem.align(self.pe.OPTIONAL_HEADER.SizeOfImage, 0x1000)

            if self.pe_image_address + self.pe_image_address_size > self.ql.os.heap_base_address:
                # pe reloc
                self.pe_image_address = self.image_address
                self.pe.relocate_image(self.image_address)

            self.entry_point = self.pe_entry_point = self.pe_image_address + self.pe.OPTIONAL_HEADER.AddressOfEntryPoint
            self.sizeOfStackReserve = self.pe.OPTIONAL_HEADER.SizeOfStackReserve
            self.ql.log.info("Loading %s to 0x%x" % (self.path, self.pe_image_address))
            self.ql.log.info("PE entry point at 0x%x" % self.entry_point)
            self.images.append(self.coverage_image(self.pe_image_address, self.pe_image_address + self.pe.NT_HEADERS.OPTIONAL_HEADER.SizeOfImage, self.path))

            # Stack should not init at the very bottom. Will cause errors with Dlls
            sp = self.stack_address + self.stack_size - 0x1000

            if self.ql.archtype == QL_ARCH.X86:
                self.ql.reg.esp = sp
                self.ql.reg.ebp = sp

                if self.pe.is_dll():
                    self.ql.log.debug('Setting up DllMain args')
                    load_addr_bytes = self.pe_image_address.to_bytes(length=4, byteorder='little')

                    self.ql.log.debug('Writing 0x%08X (IMAGE_BASE) to [ESP+4](0x%08X)' % (self.pe_image_address, sp + 0x4))
                    self.ql.mem.write(sp + 0x4, load_addr_bytes)

                    self.ql.log.debug('Writing 0x01 (DLL_PROCESS_ATTACH) to [ESP+8](0x%08X)' % (sp + 0x8))
                    self.ql.mem.write(sp + 0x8, int(1).to_bytes(length=4, byteorder='little'))

            elif self.ql.archtype == QL_ARCH.X8664:
                self.ql.reg.rsp = sp
                self.ql.reg.rbp = sp

                if self.pe.is_dll():
                    self.ql.log.debug('Setting up DllMain args')

                    self.ql.log.debug('Setting RCX (arg1) to %16X (IMAGE_BASE)' % (self.pe_image_address))
                    self.ql.reg.rcx = self.pe_image_address

                    self.ql.log.debug('Setting RDX (arg2) to 1 (DLL_PROCESS_ATTACH)')
                    self.ql.reg.rdx = 1
            else:
                raise QlErrorArch("Unknown ql.arch")

            # if this is NOT a driver, init tib/peb/ldr
            if not self.is_driver:  # userland program
                self.init_thread_information_block()
            else:   # Windows kernel driver
                super().init_driver_object()
                super().init_registry_path()
                super().init_eprocess()
                super().init_ki_user_shared_data()

                # setup IRQ Level in CR8 to PASSIVE_LEVEL (0)
                self.ql.uc.reg_write(UC_X86_REG_CR8, 0)

                # setup CR4, some drivers may check this at initialized time
                self.ql.uc.reg_write(UC_X86_REG_CR4, 0x6f8)

                self.ql.log.debug('Setting up DriverEntry args')
                self.ql.stop_execution_pattern = 0xDEADC0DE

                if self.ql.archtype == QL_ARCH.X86:  # Win32
                    if not self.ql.stop_options.any:
                        # We know that a driver will return,
                        # so if the user did not configure stop options, write a sentinel return value
                        self.ql.mem.write(sp, self.ql.stop_execution_pattern.to_bytes(length=4, byteorder='little'))

                    self.ql.log.debug('Writing 0x%08X (PDRIVER_OBJECT) to [ESP+4](0x%08X)' % (self.ql.driver_object_address, sp+0x4))
                    self.ql.log.debug('Writing 0x%08X (RegistryPath) to [ESP+8](0x%08X)' % (self.ql.regitry_path_address, sp+0x8))
                elif self.ql.archtype == QL_ARCH.X8664:  # Win64
                    if not self.ql.stop_options.any:
                        # We know that a driver will return,
                        # so if the user did not configure stop options, write a sentinel return value
                        self.ql.mem.write(sp, self.ql.stop_execution_pattern.to_bytes(length=8, byteorder='little'))

                    self.ql.log.debug('Setting RCX (arg1) to %16X (PDRIVER_OBJECT)' % (self.ql.driver_object_address))
                    self.ql.log.debug('Setting RDX (arg2) to %16X (PUNICODE_STRING)' % (self.ql.regitry_path_address))

                # setup args for DriverEntry()
                self.ql.os.set_function_args((self.ql.driver_object_address, self.ql.regitry_path_address))

            # mmap PE file into memory
            self.ql.mem.map(self.pe_image_address, self.align(self.pe_image_address_size, 0x1000), info="[PE]")
            self.pe.parse_data_directories()
            data = bytearray(self.pe.get_memory_mapped_image())
            self.ql.mem.write(self.pe_image_address, bytes(data))
            # setup IMAGE_LOAD_CONFIG_DIRECTORY
            if self.pe.OPTIONAL_HEADER.DATA_DIRECTORY[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG']].VirtualAddress != 0:
                SecurityCookie_rva = self.pe.DIRECTORY_ENTRY_LOAD_CONFIG.struct.SecurityCookie - self.pe.OPTIONAL_HEADER.ImageBase
                SecurityCookie_value = default_security_cookie_value = self.ql.mem.read(self.pe_image_address+SecurityCookie_rva, self.ql.pointersize)
                while SecurityCookie_value == default_security_cookie_value:
                    SecurityCookie_value = secrets.token_bytes(self.ql.pointersize)
                    # rol     rcx, 10h (rcx: cookie)
                    # test    cx, 0FFFFh
                    SecurityCookie_value_array = bytearray(SecurityCookie_value)
                    # Sanity question: We are always little endian, right?
                    SecurityCookie_value_array[-2:] = b'\x00\x00'
                    SecurityCookie_value = bytes(SecurityCookie_value_array)
                self.ql.mem.write(self.pe_image_address+SecurityCookie_rva, SecurityCookie_value)

            # Add main PE to ldr_data_table
            mod_name = os.path.basename(self.path)
            self.dlls[mod_name] = self.pe_image_address
            # only userland code need LDR table
            if not self.is_driver:
                super().add_ldr_data_table_entry(mod_name)

            # load system dlls
            sys_dlls = self.sys_dlls
            for each in sys_dlls:
                super().load_dll(each, self.is_driver)
            # parse directory entry import
            if self.pe.OPTIONAL_HEADER.DATA_DIRECTORY[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_IMPORT']].VirtualAddress != 0:
                for entry in self.pe.DIRECTORY_ENTRY_IMPORT:
                    dll_name = str(entry.dll.lower(), 'utf-8', 'ignore')
                    super().load_dll(entry.dll, self.is_driver)
                    for imp in entry.imports:
                        # fix IAT
                        # ql.log.info(imp.name)
                        # ql.log.info(self.import_address_table[imp.name])
                        if imp.name:
                            try:
                                addr = self.import_address_table[dll_name][imp.name]
                            except KeyError:
                                self.ql.log.debug("Error in loading function %s" % imp.name.decode())
                        else:
                            addr = self.import_address_table[dll_name][imp.ordinal]

                        if self.ql.archtype == QL_ARCH.X86:
                            address = self.ql.pack32(addr)
                        else:
                            address = self.ql.pack64(addr)
                        self.ql.mem.write(imp.address, address)

            self.ql.log.debug("Done with loading %s" % self.path)
            self.ql.os.entry_point = self.entry_point
            self.ql.os.pid = 101

        elif self.ql.code:
            self.filepath = b""
            if self.ql.archtype == QL_ARCH.X86:
                self.ql.reg.esp = self.stack_address + 0x3000
                self.ql.reg.ebp = self.ql.reg.esp
            elif self.ql.archtype == QL_ARCH.X8664:
                self.ql.reg.rsp = self.stack_address + 0x3000
                self.ql.reg.rbp = self.ql.reg.rsp

            # load shellcode in
            self.ql.mem.map(self.entry_point, self.ql.os.code_ram_size, info="[shellcode_base]")
            # rewrite entrypoint for windows shellcode
            self.ql.os.entry_point = self.entry_point
            self.ql.os.pid = 101

            self.ql.mem.write(self.entry_point, self.ql.code)
            
            self.init_thread_information_block()
            # load dlls
            for each in self.init_dlls:
                super().load_dll(each)

        # move entry_point to ql.os
        self.ql.os.entry_point = self.entry_point
        self.init_sp = self.ql.reg.arch_sp



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_sysinfo(ql, sysinfo_info, *args, **kw):

    data = b''
    data += struct.pack("QQQQQQQQQQHQQI",
                       0x1234, # uptime
                       0x2000, # loads (1 min)
                       0x2000, # loads (5 min)
                       0x2000, # loads (15 min)
                       0x10000000, # total ram
                       0x10000000, # free ram
                       0x10000000, # shared memory
                       0x0, # memory used by buffers
                       0x0, # total swap
                       0x0, # free swap
                       0x1, # nb current processes
                       0x0, # total high mem
                       0x0, # available high mem
                       0x1, # memory unit size
    )

    regreturn = 0
    #ql.mem.write(sysinfo_info, data)
    return regreturn
